/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer;

import java.io.EOFException;
import java.io.IOException;
import javax.imageio.stream.ImageInputStream;

/**
 * Decoder for the header of the MGH volume format.
 *
 * @version 8 April 2011
 */
public class MghVolumeHeaderDecoder
{
    /** Decoded header. */
    private MghVolumeHeader _header;

    /**
     * Constructs an MGH Volume Header Decoder.
     *
     * @param imageStream Input stream to the header.
     *
     * @throws IOException If an I/O error occurs.
     */
    public MghVolumeHeaderDecoder(ImageInputStream imageStream)
	throws IOException
    {
	// Read the main header elements
	_header = getMainElements(imageStream);
	int width = _header.getWidth();
	int height = _header.getHeight();
	int depth = _header.getDepth();
	int nframes = _header.getNframes();
	int type = _header.getType();

	// Use the data type to determine the number of bytes per voxel
	int MRI_UCHAR = 0;
	int MRI_INT = 1;
	int MRI_FLOAT = 3;
	int MRI_SHORT = 4;
	int MRI_TENSOR = 6;

	int bytesPerVoxel = 2;
	if (type == MRI_FLOAT) { bytesPerVoxel = 4; }
	if (type == MRI_UCHAR) { bytesPerVoxel = 1; }
	if (type == MRI_SHORT) { bytesPerVoxel = 2; }
	if (type == MRI_INT) { bytesPerVoxel = 4; }
	if (type == MRI_TENSOR) { bytesPerVoxel = 4; }

	// Skip past the image data (header = 284 bytes)
	long dataLength = nframes*width*height*depth*bytesPerVoxel;
	imageStream.seek(284 + dataLength);

	// Read acquisition information
	addAcquistionInfo(imageStream, _header);
    }

    /**
     * Gets the decoded header.
     *
     * @return Decoded header.
     */
    public MghVolumeHeader getHeader()
    {
	return _header;
    }

    /**
     * Decodes the main header elements.
     *
     * @param imageStream Input stream to the header.
     *
     * @return Main header elements.
     *
     * @throws IOException If an I/O error occurs.
     */
    public static MghVolumeHeader getMainElements(ImageInputStream imageStream)
	throws IOException
    {
	// Read the main header elements
	int version = imageStream.readInt();
	int width = imageStream.readInt();
	int height = imageStream.readInt();
	int depth = imageStream.readInt();
	int nframes = imageStream.readInt();
	int type = imageStream.readInt();
	int dof = imageStream.readInt();

	// RAS header elements exist
	short goodRasFlag = imageStream.readShort();
	if (goodRasFlag > 0) {
	    float spacingX = imageStream.readFloat();
	    float spacingY = imageStream.readFloat();
	    float spacingZ = imageStream.readFloat();

	    float xR = imageStream.readFloat();
	    float xA = imageStream.readFloat();
	    float xS = imageStream.readFloat();

	    float yR = imageStream.readFloat();
	    float yA = imageStream.readFloat();
	    float yS = imageStream.readFloat();

	    float zR = imageStream.readFloat();
	    float zA = imageStream.readFloat();
	    float zS = imageStream.readFloat();

	    float cR = imageStream.readFloat();
	    float cA = imageStream.readFloat();
	    float cS = imageStream.readFloat();

	    // Create the MGH Header
	    return new MghRasVolumeHeader(version, width, height, depth,
					  nframes, type, dof, spacingX,
					  spacingY, spacingZ, xR, xA, xS,
					  yR, yA, yS, zR, zA, zS, cR, cA,
					  cS);
	}

	// No header elements exist
	return new MghVolumeHeader(version, width, height, depth, nframes, type,
				   dof);
    }

    /**
     * Adds acquisition information to the header.
     *
     * @param imageStream Input stream to the header.
     * @param header Header to add information to.
     *
     * @throws IOException If an I/O error occurs.
     */
    public static void addAcquistionInfo(ImageInputStream imageStream,
					 MghVolumeHeader header)
	throws IOException
    {
	// Read acquisition information
	try {
	    float tr = imageStream.readFloat();
	    try {
		float flipAngle = imageStream.readFloat();
		try {
		    float te = imageStream.readFloat();
		    try {
			float ti = imageStream.readFloat();
			try {
			    float fov = imageStream.readFloat();
			    header.setFloats(tr, flipAngle, te, ti, fov);
			}
			catch (EOFException eof) {
			    header.setFloats(tr, flipAngle, te, ti);
			    return;
			}
		    }
		    catch (EOFException eof) {
			header.setFloats(tr, flipAngle, te);
			return;
		    }
		}
		catch (EOFException eof) {
		    header.setFloats(tr, flipAngle);
		    return;
		}
	    }
	    catch (EOFException eof) {
		header.setFloats(tr);
		return;
	    }
	}
	catch (EOFException eof) { return; }

	// Read tags
	try {
	    while (true) {
		int tagCode = imageStream.readInt();

		// Determine the tag length
		long length;
		if (tagCode == MghTag.OLD_MGH_XFORM) {
		    length = imageStream.readInt();
		}
		else if (tagCode == MghTag.OLD_SURF_GEOM ||
			 tagCode == MghTag.OLD_USEREALRAS ||
			 tagCode == MghTag.OLD_COLORTABLE)
		    {
			length = 0;
		    }
		else { length = imageStream.readLong(); }

		// Read the tag value
		byte[] b = new byte[(int)length];
		imageStream.readFully(b);
		String tagValue = new String(b);

		// Add the tag
		header.addTag(tagCode, tagValue);
	    }
	}
	catch (EOFException eof) {}
    }
}
