/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer;

import java.io.DataOutput;
import java.io.IOException;

/**
 * Encoder for the header of the MGH volume format.
 *
 * @version 8 April 2011
 */
public class MghVolumeHeaderEncoder
{
    /** Output stream for the header. */
    private DataOutput _imageStream;

    /**
     * Constructs an MGH Volume Header Encoder.
     *
     * @param imageStream Output stream for the header.
     */
    public MghVolumeHeaderEncoder(DataOutput imageStream) 
    {
	_imageStream = imageStream;
    }

    /**
     * Writes the main elements of the header.
     *
     * @param header Header to encode.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void writeMainElements(MghVolumeHeader header) throws IOException
    {
	// Write the main header elements
	_imageStream.writeInt( header.getVersion() );
	_imageStream.writeInt( header.getWidth() );
	_imageStream.writeInt( header.getHeight() );
	_imageStream.writeInt( header.getDepth() );
	_imageStream.writeInt( header.getNframes() );
	_imageStream.writeInt( header.getType() );
	_imageStream.writeInt( header.getDof() );

	// Write the RAS elements
	if (header instanceof MghRasVolumeHeader) {
	    MghRasVolumeHeader rasHeader = (MghRasVolumeHeader)header;

	    _imageStream.writeShort(1);
	    _imageStream.writeFloat( rasHeader.getSpacingX() );
	    _imageStream.writeFloat( rasHeader.getSpacingY() );
	    _imageStream.writeFloat( rasHeader.getSpacingZ() );
	    _imageStream.writeFloat( rasHeader.getXr() );
	    _imageStream.writeFloat( rasHeader.getXa() );
	    _imageStream.writeFloat( rasHeader.getXs() );
	    _imageStream.writeFloat( rasHeader.getYr() );
	    _imageStream.writeFloat( rasHeader.getYa() );
	    _imageStream.writeFloat( rasHeader.getYs() );
	    _imageStream.writeFloat( rasHeader.getZr() );
	    _imageStream.writeFloat( rasHeader.getZa() );
	    _imageStream.writeFloat( rasHeader.getZs() );
	    _imageStream.writeFloat( rasHeader.getCr() );
	    _imageStream.writeFloat( rasHeader.getCa() );
	    _imageStream.writeFloat( rasHeader.getCs() );
	}
	else { _imageStream.writeShort(-1); }

	// Write padding
	byte[] zeroes = new byte[194];
	_imageStream.write(zeroes);
    }

    /**
     * Writes the acquisition information of the header.
     *
     * @param header Header to encode.
     *
     * @throws IOException If an I/O error occurs.
     */
    public void writeAcquistionInfo(MghVolumeHeader header) throws IOException
    {
	// Write the float values
	MghFloat[] floats = header.getFloats();
	for (int i = 0; i < floats.length; i++) {
	    _imageStream.writeFloat( floats[i].getValue() );
	}

	// Write tags
	MghTag[] tags = header.getTags();
	for (int i = 0; i < tags.length; i++) {

	    // Write the tag code
	    int code = tags[i].getCode();
	    _imageStream.writeInt(code);

	    // Get the tag value
	    byte[] valueBytes = tags[i].getValue().getBytes();

	    // Write the tag length
	    if (code == MghTag.OLD_MGH_XFORM) {
		_imageStream.writeInt(valueBytes.length);
	    }
	    else if (code != MghTag.OLD_SURF_GEOM &&
		     code != MghTag.OLD_USEREALRAS &&
		     code != MghTag.OLD_COLORTABLE)
		{
		    _imageStream.writeLong(valueBytes.length);
		}

	    // Write the tag value
	    _imageStream.write(valueBytes);
	}
    }
}
