/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.freesurfer.plugin;

import edu.ucla.loni.Conversions;
import edu.ucla.loni.freesurfer.AsciiCurvMeta;
import edu.ucla.loni.freesurfer.CurvMeta;
import edu.ucla.loni.freesurfer.MghFloat;
import edu.ucla.loni.freesurfer.MghRasVolumeHeader;
import edu.ucla.loni.freesurfer.MghTag;
import edu.ucla.loni.freesurfer.MghVolumeHeader;
import edu.ucla.loni.freesurfer.TriangleSurfaceMeta;
import edu.ucla.loni.freesurfer.XfmMeta;
import java.util.Vector;
import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadataNode;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Class that provides conversion between Free Surfer metadata and the DOM Node
 * representations of it.
 *
 * @version 14 January 2010
 */
public class FreeSurferMetadataConversions
{
    /** Constructs a Free Surfer Metadata Conversions. */
    private FreeSurferMetadataConversions()
    {
    }

    /**
     * Converts the tree to an MGH Volume Header.
     *
     * @param root DOM Node representing the MGH Volume Header.
     *
     * @return MGH Volume Header constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     * @throws IllegalArgumentException If the root Node is null.
     */
    public static MghVolumeHeader toMghVolumeHeader(Node root)
	throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check that the root Node name matches the required name
	String nm = MghMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( !root.getNodeName().equals(nm) ) {
	    String msg = "Root node must be named \"" + nm + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Attempt to construct the MGH Volume Header
	try {
	    NamedNodeMap atts = root.getAttributes();

	    // Version
	    String id = MghMetadataFormat.VERSION;
	    int versn = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Width
	    id = MghMetadataFormat.WIDTH;
	    int width = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Height
	    id = MghMetadataFormat.HEIGHT;
	    int height = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Depth
	    id = MghMetadataFormat.DEPTH;
	    int depth = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Nframes
	    id = MghMetadataFormat.NFRAMES;
	    int nfrms = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Type
	    id = MghMetadataFormat.TYPE;
	    int type = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Dof
	    id = MghMetadataFormat.DOF;
	    int dof = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Construct the Header
	    MghVolumeHeader header;
	    if (atts.getNamedItem(MghMetadataFormat.SPACING_X) != null) {

		// Spacing X
		id = MghMetadataFormat.SPACING_X;
		float sX=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Spacing Y
		id = MghMetadataFormat.SPACING_Y;
		float sY=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Spacing Z
		id = MghMetadataFormat.SPACING_Z;
		float sZ=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Xr
		id = MghMetadataFormat.XR;
		float xR=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Xa
		id = MghMetadataFormat.XA;
		float xA=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Xs
		id = MghMetadataFormat.XS;
		float xS=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Yr
		id = MghMetadataFormat.YR;
		float yR=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Ya
		id = MghMetadataFormat.YA;
		float yA=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Ys
		id = MghMetadataFormat.YS;
		float yS=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Zr
		id = MghMetadataFormat.ZR;
		float zR=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Za
		id = MghMetadataFormat.ZA;
		float zA=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Zs
		id = MghMetadataFormat.ZS;
		float zS=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Cr
		id = MghMetadataFormat.CR;
		float cR=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Ca
		id = MghMetadataFormat.CA;
		float cA=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Cs
		id = MghMetadataFormat.CS;
		float cS=Float.parseFloat(atts.getNamedItem(id).getNodeValue());

		// Construct the header
		header = new MghRasVolumeHeader(versn, width, height, depth,
						nfrms, type, dof, sX, sY, sZ,
						xR, xA, xS, yR, yA, yS, zR, zA,
						zS, cR, cA,cS);
	    }
	    else {
		header = new MghVolumeHeader(versn, width, height, depth, nfrms,
					     type, dof);
	    }

	    // Look for more information
	    Node childNode = root.getFirstChild();
	    if (childNode == null) { return header; }

	    // Add acquisition information
	    if (MghMetadataFormat.ACQUISITION.equals(childNode.getNodeName())) {
		_addFloatValues(header, childNode);
	    }

	    // Add optional tags
	    NodeList childNodes = root.getChildNodes();
	    for (int i = 0; i < childNodes.getLength(); i++) {
		childNode = childNodes.item(i);

		// Found a tag
		if ( MghMetadataFormat.TAG.equals(childNode.getNodeName()) ) {
		    atts = childNode.getAttributes();

		    // Tag code
		    id = MghMetadataFormat.TAG_CODE;
		    int code =
			Integer.parseInt(atts.getNamedItem(id).getNodeValue());

		    // Tag value
		    id = MghMetadataFormat.TAG_VALUE;
		    String value = atts.getNamedItem(id).getNodeValue();

		    // Add the tag information
		    header.addTag(code, value);
		}
	    }

	    // Return the Header
	    return header;
	}

	// Unable to construct the MGH Volume Header
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into an MGH Volume " +
		"Header.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Adds float values to the MGH Volume Header.
     *
     * @param header Hear to add float values to.
     * @param floatNode Node of float values.
     */
    private static void _addFloatValues(MghVolumeHeader header, Node floatNode)
    {
	NamedNodeMap atts = floatNode.getAttributes();

	// TR
	Node attr = atts.getNamedItem(MghFloat.TR);
	if (attr == null) { return; }
	float tr = Float.parseFloat(attr.getNodeValue());

	// Flip angle
	attr = atts.getNamedItem(MghFloat.FLIP_ANGLE);
	if (attr == null) {
	    header.setFloats(tr);
	    return;
	}
	float angle = Float.parseFloat(attr.getNodeValue());

	// TE
	attr = atts.getNamedItem(MghFloat.TE);
	if (attr == null) {
	    header.setFloats(tr, angle);
	    return;
	}
	float te = Float.parseFloat(attr.getNodeValue());

	// TI
	attr = atts.getNamedItem(MghFloat.TI);
	if (attr == null) {
	    header.setFloats(tr, angle, te);
	    return;
	}
	float ti = Float.parseFloat(attr.getNodeValue());

	// FOV
	attr = atts.getNamedItem(MghFloat.FOV);
	if (attr == null) {
	    header.setFloats(tr, angle, te, ti);
	    return;
	}
	float fov = Float.parseFloat(attr.getNodeValue());

	// Set the header information
	header.setFloats(tr, angle, te, ti, fov);
    }

    /**
     * Converts the tree to an ASCII Curv Meta.
     *
     * @param root DOM Node representing the ASCII Curv Meta.
     *
     * @return ASCII Curv Meta constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     * @throws IllegalArgumentException If the root Node is null.
     */
    public static AsciiCurvMeta toAsciiCurvMeta(Node root)
	throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check that the root Node name matches the required name
	String nm = AsciiCurvFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( !root.getNodeName().equals(nm) ) {
	    String msg = "Root node must be named \"" + nm + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Attempt to construct the ASCII Curv Meta
	try {
	    NamedNodeMap atts = root.getAttributes();

	    // Comment
	    String id = AsciiCurvFormat.COMMENT;
	    byte[] comment = atts.getNamedItem(id).getNodeValue().getBytes();

	    // Vertex count
	    id = AsciiCurvFormat.VERTEX_COUNT;
	    int vCnt = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Construct the ASCII Curv Meta
	    return new AsciiCurvMeta(comment, vCnt);
	}

	// Unable to construct the ASCII Curv Meta
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into an ASCII Curv Meta.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the tree to a Curv Meta.
     *
     * @param root DOM Node representing the Curv Meta.
     *
     * @return Curv Meta constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     * @throws IllegalArgumentException If the root Node is null.
     */
    public static CurvMeta toCurvMeta(Node root) throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check that the root Node name matches the required name
	String nm = CurvFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( !root.getNodeName().equals(nm) ) {
	    String msg = "Root node must be named \"" + nm + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Attempt to construct the Curv Meta
	try {
	    NamedNodeMap atts = root.getAttributes();

	    // Vertex count
	    String id = CurvFormat.VERTEX_COUNT;
	    int vCnt = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Face count
	    id = CurvFormat.FACE_COUNT;
	    int fCnt = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Values per vertex
	    id = CurvFormat.VALS_PER_VERTEX;
	    int vpv = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Construct the Curv Meta
	    return new CurvMeta(vCnt, fCnt, vpv);
	}

	// Unable to construct the Curv Meta
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into a Curv Meta.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the tree to a Triangle Surface Meta.
     *
     * @param root DOM Node representing the Triangle Surface Meta.
     *
     * @return Triangle Surface Meta constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     * @throws IllegalArgumentException If the root Node is null.
     */
    public static TriangleSurfaceMeta toTriangleSurfaceMeta(Node root)
	throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check that the root Node name matches the required name
	String nm = TriangleSurfaceFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( !root.getNodeName().equals(nm) ) {
	    String msg = "Root node must be named \"" + nm + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Attempt to construct the Triangle Surface Meta
	try {
	    NamedNodeMap atts = root.getAttributes();

	    // Creator/user
	    String id = TriangleSurfaceFormat.CREATOR_USER;
	    byte[] createUser = atts.getNamedItem(id).getNodeValue().getBytes();

	    // Vertex count
	    id = TriangleSurfaceFormat.VERTEX_COUNT;
	    int vCnt = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Face count
	    id = TriangleSurfaceFormat.FACE_COUNT;
	    int fCnt = Integer.parseInt(atts.getNamedItem(id).getNodeValue());

	    // Footer
	    id = TriangleSurfaceFormat.FOOTER;
	    String footerStr = atts.getNamedItem(id).getNodeValue();
	    byte[] footer = Conversions.toByteArray(footerStr);

	    // Return the Triangle Surface Meta
	    return new TriangleSurfaceMeta(createUser, vCnt, fCnt, footer);
	}

	// Unable to construct the Triangle Surface Meta
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into a Triangle Surface "+
		"Meta.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the tree to an XFM Meta.
     *
     * @param root DOM Node representing the XFM Meta.
     *
     * @return XFM Meta constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     * @throws IllegalArgumentException If the root Node is null.
     */
    public static XfmMeta toXfmMeta(Node root) throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check that the root Node name matches the required name
	String nm = XfmFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( !root.getNodeName().equals(nm) ) {
	    String msg = "Root node must be named \"" + nm + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Attempt to construct the XFM Meta
	try {
	    NamedNodeMap atts = root.getAttributes();

	    // Transform type
	    String id = XfmFormat.TRANSFORM_TYPE;
	    String transformType = atts.getNamedItem(id).getNodeValue();

	    // Comments
	    Vector comments = new Vector();

	    // Comment Nodes
	    NodeList childNodes = root.getChildNodes();
	    for (int i = 0; i < childNodes.getLength(); i++) {
		Node childNode = childNodes.item(i);

		// Found a comment
		if ( XfmFormat.COMMENT.equals(childNode.getNodeName()) ) {
		    atts = childNode.getAttributes();
		    id = XfmFormat.COMMENT_TEXT;
		    comments.add(atts.getNamedItem(id).getNodeValue());
		}
	    }

	    // Linear transform Node
	    Node linearNode = null;
	    for (int i = 0; i < childNodes.getLength(); i++) {
		Node childNode = childNodes.item(i);
		if (XfmFormat.LINEAR_TRANSFORM.equals(childNode.getNodeName())){
		    linearNode = childNode;
		}
	    }
	    atts = linearNode.getAttributes();

	    // Linear transform elements
	    String[] elems = new String[12];
	    elems[0] = atts.getNamedItem(XfmFormat.ELEM_1_1).getNodeValue();
	    elems[1] = atts.getNamedItem(XfmFormat.ELEM_1_2).getNodeValue();
	    elems[2] = atts.getNamedItem(XfmFormat.ELEM_1_3).getNodeValue();
	    elems[3] = atts.getNamedItem(XfmFormat.ELEM_1_4).getNodeValue();
	    elems[4] = atts.getNamedItem(XfmFormat.ELEM_2_1).getNodeValue();
	    elems[5] = atts.getNamedItem(XfmFormat.ELEM_2_2).getNodeValue();
	    elems[6] = atts.getNamedItem(XfmFormat.ELEM_2_3).getNodeValue();
	    elems[7] = atts.getNamedItem(XfmFormat.ELEM_2_4).getNodeValue();
	    elems[8] = atts.getNamedItem(XfmFormat.ELEM_3_1).getNodeValue();
	    elems[9] = atts.getNamedItem(XfmFormat.ELEM_3_2).getNodeValue();
	    elems[10] = atts.getNamedItem(XfmFormat.ELEM_3_3).getNodeValue();
	    elems[11] = atts.getNamedItem(XfmFormat.ELEM_3_4).getNodeValue();

	    // Construct the XFM Meta
	    return new XfmMeta(transformType,
			       (String[])comments.toArray(new String[0]),
			       elems);
	}

	// Unable to construct the XFM Meta
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into an XFM Meta.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the MGH Volume Header to a tree.
     *
     * @param header MGH Volume Header to convert.
     *
     * @return DOM Node representing the MGH Volume Header.
     */
    public static Node toTree(MghVolumeHeader header)
    {
      // Create the root Node
      String root = MghMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
      IIOMetadataNode rootNode = new IIOMetadataNode(root);

      // Add the attributes
      String name = MghMetadataFormat.VERSION;
      rootNode.setAttribute(name, header.getVersion()+"");

      name = MghMetadataFormat.WIDTH;
      rootNode.setAttribute(name, header.getWidth()+"");

      name = MghMetadataFormat.HEIGHT;
      rootNode.setAttribute(name, header.getHeight()+"");

      name = MghMetadataFormat.DEPTH;
      rootNode.setAttribute(name, header.getDepth()+"");

      name = MghMetadataFormat.NFRAMES;
      rootNode.setAttribute(name, header.getNframes()+"");

      name = MghMetadataFormat.TYPE;
      rootNode.setAttribute(name, header.getType()+"");

      name = MghMetadataFormat.DOF;
      rootNode.setAttribute(name, header.getDof()+"");

      if (header instanceof MghRasVolumeHeader) {
	  MghRasVolumeHeader rasHeader = (MghRasVolumeHeader)header;

	  name = MghMetadataFormat.SPACING_X;
	  rootNode.setAttribute(name, rasHeader.getSpacingX()+"");

	  name = MghMetadataFormat.SPACING_Y;
	  rootNode.setAttribute(name, rasHeader.getSpacingY()+"");

	  name = MghMetadataFormat.SPACING_Z;
	  rootNode.setAttribute(name, rasHeader.getSpacingZ()+"");

	  name = MghMetadataFormat.XR;
	  rootNode.setAttribute(name, rasHeader.getXr()+"");

	  name = MghMetadataFormat.XA;
	  rootNode.setAttribute(name, rasHeader.getXa()+"");

	  name = MghMetadataFormat.XS;
	  rootNode.setAttribute(name, rasHeader.getXs()+"");

	  name = MghMetadataFormat.YR;
	  rootNode.setAttribute(name, rasHeader.getYr()+"");

	  name = MghMetadataFormat.YA;
	  rootNode.setAttribute(name, rasHeader.getYa()+"");

	  name = MghMetadataFormat.YS;
	  rootNode.setAttribute(name, rasHeader.getYs()+"");

	  name = MghMetadataFormat.ZR;
	  rootNode.setAttribute(name, rasHeader.getZr()+"");

	  name = MghMetadataFormat.ZA;
	  rootNode.setAttribute(name, rasHeader.getZa()+"");

	  name = MghMetadataFormat.ZS;
	  rootNode.setAttribute(name, rasHeader.getZs()+"");

	  name = MghMetadataFormat.CR;
	  rootNode.setAttribute(name, rasHeader.getCr()+"");

	  name = MghMetadataFormat.CA;
	  rootNode.setAttribute(name, rasHeader.getCa()+"");

	  name = MghMetadataFormat.CS;
	  rootNode.setAttribute(name, rasHeader.getCs()+"");
      }

      // Create the acquisition node
      MghFloat[] floats = header.getFloats();
      if (floats.length > 0) {
	  name = MghMetadataFormat.ACQUISITION;
	  IIOMetadataNode floatNode = new IIOMetadataNode(name);
	  rootNode.appendChild(floatNode);

	  // Add the acquisition attributes
	  for (int i = 0; i < floats.length; i++) {
	      floatNode.setAttribute(floats[i].getName(),
				     floats[i].getValue()+"");
	  }
      }

      // Create the tag nodes
      MghTag[] tags = header.getTags();
      for (int i = 0; i < tags.length; i++) {
	  name = MghMetadataFormat.TAG;
	  IIOMetadataNode tagNode = new IIOMetadataNode(name);
	  rootNode.appendChild(tagNode);

	  // Add the tag attributes
	  tagNode.setAttribute(MghMetadataFormat.TAG_CODE,
			       tags[i].getCode()+"");
	  tagNode.setAttribute(MghMetadataFormat.TAG_VALUE,
			       tags[i].getValue());
      }

      // Return the root Node
      return rootNode;
    }

    /**
     * Converts the ASCII Curv Meta to a tree.
     *
     * @param meta ASCII Curv Meta to convert.
     *
     * @return DOM Node representing the ASCII Curv Meta.
     */
    public static Node toTree(AsciiCurvMeta meta)
    {
      // Create the root Node
      String root = AsciiCurvFormat.NATIVE_METADATA_FORMAT_NAME;
      IIOMetadataNode rootNode = new IIOMetadataNode(root);

      // Add the attributes
      String name = AsciiCurvFormat.COMMENT;
      rootNode.setAttribute(name, meta.getComment());

      name = AsciiCurvFormat.VERTEX_COUNT;
      rootNode.setAttribute(name, meta.getVertexCount()+"");

      // Return the root Node
      return rootNode;
    }

    /**
     * Converts the Curv Meta to a tree.
     *
     * @param meta Curv Meta to convert.
     *
     * @return DOM Node representing the Curv Meta.
     */
    public static Node toTree(CurvMeta meta)
    {
      // Create the root Node
      String root = CurvFormat.NATIVE_METADATA_FORMAT_NAME;
      IIOMetadataNode rootNode = new IIOMetadataNode(root);

      // Add the attributes
      String name = CurvFormat.VERTEX_COUNT;
      rootNode.setAttribute(name, meta.getVertexCount()+"");

      name = CurvFormat.FACE_COUNT;
      rootNode.setAttribute(name, meta.getFaceCount()+"");

      name = CurvFormat.VALS_PER_VERTEX;
      rootNode.setAttribute(name, meta.getValsPerVertex()+"");

      // Return the root Node
      return rootNode;
    }

    /**
     * Converts the Triangle Surface Meta to a tree.
     *
     * @param meta Triangle Surface Meta to convert.
     *
     * @return DOM Node representing the Triangle Surface Meta.
     */
    public static Node toTree(TriangleSurfaceMeta meta)
    {
      // Create the root Node
      String root = TriangleSurfaceFormat.NATIVE_METADATA_FORMAT_NAME;
      IIOMetadataNode rootNode = new IIOMetadataNode(root);

      // Add the attributes
      String name = TriangleSurfaceFormat.CREATOR_USER;
      rootNode.setAttribute(name, meta.getCreatorUser());

      name = TriangleSurfaceFormat.VERTEX_COUNT;
      rootNode.setAttribute(name, meta.getVertexCount()+"");

      name = TriangleSurfaceFormat.FACE_COUNT;
      rootNode.setAttribute(name, meta.getFaceCount()+"");

      name = TriangleSurfaceFormat.FOOTER;
      rootNode.setAttribute(name, Conversions.toHexString(meta.getFooter()));

      // Return the root Node
      return rootNode;
    }

    /**
     * Converts the XFM Meta to a tree.
     *
     * @param meta XFM Meta to convert.
     *
     * @return DOM Node representing the XFM Meta.
     */
    public static Node toTree(XfmMeta meta)
    {
      // Create the root Node
      String root = XfmFormat.NATIVE_METADATA_FORMAT_NAME;
      IIOMetadataNode rootNode = new IIOMetadataNode(root);

      // Add the attribute
      String name = XfmFormat.TRANSFORM_TYPE;
      rootNode.setAttribute(name, meta.getTransformType());

      // Add comments
      String[] comments = meta.getComments();
      for (int i = 0; i < comments.length; i++) {

	  // Add a comment Node
	  name = XfmFormat.COMMENT;
	  IIOMetadataNode commentNode = new IIOMetadataNode(name);
	  rootNode.appendChild(commentNode);

	  // Add the comment text
	  commentNode.setAttribute(XfmFormat.COMMENT_TEXT, comments[i]);
      }

      // Add the linear transform Node
      name = XfmFormat.LINEAR_TRANSFORM;
      IIOMetadataNode linearNode = new IIOMetadataNode(name);
      rootNode.appendChild(linearNode);

      // Add the linear transform attributes
      linearNode.setAttribute(XfmFormat.ELEM_1_1, meta.getTransformElem(0));
      linearNode.setAttribute(XfmFormat.ELEM_1_2, meta.getTransformElem(1));
      linearNode.setAttribute(XfmFormat.ELEM_1_3, meta.getTransformElem(2));
      linearNode.setAttribute(XfmFormat.ELEM_1_4, meta.getTransformElem(3));
      linearNode.setAttribute(XfmFormat.ELEM_2_1, meta.getTransformElem(4));
      linearNode.setAttribute(XfmFormat.ELEM_2_2, meta.getTransformElem(5));
      linearNode.setAttribute(XfmFormat.ELEM_2_3, meta.getTransformElem(6));
      linearNode.setAttribute(XfmFormat.ELEM_2_4, meta.getTransformElem(7));
      linearNode.setAttribute(XfmFormat.ELEM_3_1, meta.getTransformElem(8));
      linearNode.setAttribute(XfmFormat.ELEM_3_2, meta.getTransformElem(9));
      linearNode.setAttribute(XfmFormat.ELEM_3_3, meta.getTransformElem(10));
      linearNode.setAttribute(XfmFormat.ELEM_3_4, meta.getTransformElem(11));

      // Return the root Node
      return rootNode;
    }
}
