/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom;

/**
 * Data Dictionary of private DICOM Data Elements for GE scanners.
 *
 * @version 23 February 2011
 */
public class GeDataDictionary extends DataDictionary
{
    /** Single instance of the GE Data Dictionary. */
    private static DataDictionary _dictionary = new GeDataDictionary();

    /** Constructs a GE Data Dictionary. */
    private GeDataDictionary()
    {
    }

    /**
     * Gets the instance of the GE Data Dictionary.
     *
     * @return Single instance of the GE Data Dictionary.
     */
    public static DataDictionary getInstance()
    {
	return _dictionary;
    }

    /**
     * Determines if this Data Dictionary has private Data Elements from the
     * manufacturer.
     *
     * @param manufacturer Manufacturer.
     *
     * @return True if there are private Data Elements; false otherwise.
     */
    public static boolean isDictionaryFor(String manufacturer)
    {
	if (manufacturer == null) { return false; }
	return manufacturer.trim().startsWith("GE");
    }

    /** Loads the registry of DICOM Data Elements. */
    protected void _loadRegistry()
    {
	// Group 0009 Elements
	_add("00090000", UL, "Group Length", VM_1);
	_add("00090010", LO, "GEHC Private Creator ID", VM_1);
	_add("00091001", LO, "Full fidelity", VM_1);
	_add("00091002", SH, "Suite id", VM_1);
	_add("00091004", SH, "Product id", VM_1);
	_add("00091027", SL, "Image actual date", VM_1);
	_add("00091030", SH, "Service id", VM_1);
	_add("00091031", SH, "Mobile location number", VM_1);
	_add("000910E3", UI, "Equipment UID", VM_1);
	_add("000910E6", SH, "Genesis Version - now ", VM_1);
	_add("000910E7", UL, "Exam Record checksum", VM_1);
	_add("000910E9", SL, "Actual series data time stamp", VM_1);

	// Group 0011 Elements
	_add("00110000", UL, "Group Length", VM_1);
	_add("00110010", LO, "GEHC Private Creator ID", VM_1);
	_add("00111010", SS, "Patient Status", VM_1);

	// Group 0019 Elements
	_add("00190000", UL, "Group Length", VM_1);
	_add("00190010", LO, "GEHC Private Creator ID", VM_1);
	_add("0019100F", DS, "Horiz. Frame of ref.", VM_1);
	_add("00191011", SS, "Series contrast", VM_1);
	_add("00191012", SS, "Last pseq", VM_1);
	_add("00191017", SS, "Series plane", VM_1);
	_add("00191018", LO, "First scan ras", VM_1);
	_add("00191019", DS, "First scan location", VM_1);
	_add("0019101A", LO, "Last scan ras", VM_1);
	_add("0019101B", DS, "Last scan loc", VM_1);
	_add("0019101E", DS, "Display field of view", VM_1);
	_add("0019105A", FL, "Acquisition Duration", VM_1);
	_add("0019107D", DS, "Second echo", VM_1);
	_add("0019107E", SS, "Number of echoes", VM_1);
	_add("0019107F", DS, "Table delta", VM_1);
	_add("00191081", SS, "Contiguous", VM_1);
	_add("00191084", DS, "Peak SAR", VM_1);
	_add("00191085", SS, "Monitor SAR", VM_1);
	_add("00191087", DS, "Cardiac repetition time", VM_1);
	_add("00191088", SS, "Images per cardiac cycle", VM_1);
	_add("0019108A", SS, "Actual receive gain analog", VM_1);
	_add("0019108B", SS, "Actual receive gain digital", VM_1);
	_add("0019108D", DS, "Delay after trigger", VM_1);
	_add("0019108F", SS, "Swappf", VM_1);
	_add("00191090", SS, "Pause Interval", VM_1);
	_add("00191091", DS, "Pause Time", VM_1);
	_add("00191092", SL, "Slice offset on freq axis", VM_1);
	_add("00191093", DS, "Auto Prescan Center Frequency", VM_1);
	_add("00191094", SS, "Auto Prescan Transmit Gain", VM_1);
	_add("00191095", SS, "Auto Prescan Analog receiver gain", VM_1);
	_add("00191096", SS, "Auto Prescan Digital receiver gain", VM_1);
	_add("00191097", SL, "Bitmap defining CVs", VM_1);
	_add("00191098", SS, "Center freq. Method", VM_1);
	_add("0019109B", SS, "Pulse Sequence Mode", VM_1);
	_add("0019109C", LO, "Pulse Sequence Name", VM_1);
	_add("0019109D", DT, "Pulse Sequence Date", VM_1);
	_add("0019109E", LO, "Internal Pulse Sequence Name", VM_1);
	_add("0019109F", SS, "Transmitting Coil Type", VM_1);
	_add("001910A0", SS, "Surface Coil Type", VM_1);
	_add("001910A1", SS, "Extremity Coil flag", VM_1);
	_add("001910A2", SL, "Raw data run number", VM_1);
	_add("001910A3", UL, "Calibrated Field strength", VM_1);
	_add("001910A4", SS, "SAT fat/water/bone", VM_1);
	_add("001910A5", DS, "Receive bandwidth", VM_1);
	_add("001910A7", DS, "User data 0", VM_1);
	_add("001910A8", DS, "User data 1", VM_1);
	_add("001910A9", DS, "User data 2", VM_1);
	_add("001910AA", DS, "User data 3", VM_1);
	_add("001910AB", DS, "User data 4", VM_1);
	_add("001910AC", DS, "User data 5", VM_1);
	_add("001910AD", DS, "User data 6", VM_1);
	_add("001910AE", DS, "User data 7", VM_1);
	_add("001910AF", DS, "User data 8", VM_1);
	_add("001910B0", DS, "User data 9", VM_1);
	_add("001910B1", DS, "User data 10", VM_1);
	_add("001910B2", DS, "User data 11", VM_1);
	_add("001910B3", DS, "User data 12", VM_1);
	_add("001910B4", DS, "User data 13", VM_1);
	_add("001910B5", DS, "User data 14", VM_1);
	_add("001910B6", DS, "User data 15", VM_1);
	_add("001910B7", DS, "User data 16", VM_1);
	_add("001910B8", DS, "User data 17", VM_1);
	_add("001910B9", DS, "User data 18", VM_1);
	_add("001910BA", DS, "User data 19", VM_1);
	_add("001910BB", DS, "User data 20", VM_1);
	_add("001910BC", DS, "User data 21", VM_1);
	_add("001910BD", DS, "User data 22", VM_1);
	_add("001910BE", DS, "Projection angle", VM_1);
	_add("001910C0", SS, "Saturation planes", VM_1);
	_add("001910C1", SS, "Surface coil intensity", VM_1);
	_add("001910C2", SS, "SAT location R", VM_1);
	_add("001910C3", SS, "SAT location L", VM_1);
	_add("001910C4", SS, "SAT location A", VM_1);
	_add("001910C5", SS, "SAT location P", VM_1);
	_add("001910C6", SS, "SAT location H", VM_1);
	_add("001910C7", SS, "SAT location F", VM_1);
	_add("001910C8", SS, "SAT thickness R/L", VM_1);
	_add("001910C9", SS, "SAT thickness A/P", VM_1);
	_add("001910CA", SS, "SAT thickness H/F", VM_1);
	_add("001910CB", SS, "Phase Contrast flow axis", VM_1);
	_add("001910CC", SS, "Velocity encoding", VM_1);
	_add("001910CD", SS, "Thickness disclaimer", VM_1);
	_add("001910CE", SS, "Prescan type", VM_1);
	_add("001910CF", SS, "Prescan status", VM_1);
	_add("001910D0", SH, "Raw data type", VM_1);
	_add("001910D2", SS, "Projection Algorithm", VM_1);
	_add("001910D3", SH, "Projection Algorithm Name", VM_1);
	_add("001910D5", SS, "Fractional echo", VM_1);
	_add("001910D6", SS, "Prep pulse", VM_1);
	_add("001910D7", SS, "Cardiac phase number", VM_1);
	_add("001910D8", SS, "Variable echoflag", VM_1);
	_add("001910D9", DS,
	     "Concatenated SAT {# DTI Diffusion Dir., release 9.0 & below}",
	     VM_1);
	_add("001910DF", DS,
	     "User data 23 {# DTI Diffusion Dir., release 9.0 & below}", VM_1);
	_add("001910E0", DS,
	     "User data 24 {# DTI Diffusion Dir., release 10.0 & above}", VM_1);
	_add("001910E2", DS, "Velocity Encode Scale", VM_1);
	_add("001910F2", DS, "Fast phases", VM_1);
	_add("001910F9", DS, "Transmit gain", VM_1);

	// Group 0021 Elements
	_add("00210000", UL, "Group Length", VM_1);
	_add("00210010", LO, "GEHC Private Creator ID", VM_1);
	_add("00211003", SS, "Series from which Prescribed", VM_1);
	_add("00211005", SH, "Genesis Version ~ now", VM_1);
	_add("00211007", UL, "Series Record checksum", VM_1);
	_add("00211018", SH, "Genesis version ~ Now", VM_1);
	_add("00211019", UL, "Acq recon record checksum", VM_1);
	_add("00211020", DS, "Table start location", VM_1);
	_add("00211035", SS, "Series from which prescribed", VM_1);
	_add("00211036", SS, "Image from which prescribed", VM_1);
	_add("00211037", SS, "Screen Format", VM_1);
	_add("0021104F", SS, "Locations in acquisition", VM_1);
	_add("00211050", SS, "Graphically prescribed", VM_1);
	_add("00211051", DS, "Rotation from source x rot", VM_1);
	_add("00211052", DS, "Rotation from source y rot", VM_1);
	_add("00211053", DS, "Rotation from source z rot", VM_1);
	_add("00211054", SH, "Image position", "3");
	_add("00211055", SH, "Image orientation", "6");
	_add("00211056", SL, "Num 3D slabs", VM_1);
	_add("00211057", SL, "Locs per 3D slab", VM_1);
	_add("00211058", SL, "Overlaps", VM_1);
	_add("00211059", SL, "Image Filtering 0.5/0.2T", VM_1);
	_add("0021105A", SL, "Diffusion direction", VM_1);
	_add("0021105B", DS, "Tagging Flip Angle", VM_1);
	_add("0021105C", DS, "Tagging Orientation", VM_1);
	_add("0021105D", DS, "Tag Spacing", VM_1);
	_add("0021105E", DS, "RTIA_timer", VM_1);
	_add("0021105F", DS, "Fps", VM_1);
	_add("00211081", DS, "Auto window/level alpha", VM_1);
	_add("00211082", DS, "Auto window/level beta", VM_1);
	_add("00211083", DS, "Auto window/level window", VM_1);
	_add("00211084", DS, "Auto window/level level", VM_1);

	// Group 0023 Elements
	_add("00230000", UL, "Group Length", VM_1);
	_add("00230010", LO, "GEHC Private Creator ID", VM_1);
	_add("00231001", SL, "Number of series in Study", VM_1);
	_add("00231002", SL, "Number of unarchived Series", VM_1);
	_add("00231010", SS, "Reference image field", VM_1);
	_add("00231050", SS, "Summary image", VM_1);
	_add("00231070", FD, "Start time(secs) in first axial", VM_1);
	_add("00231074", SL, "No. of updates to header", VM_1);
	_add("0023107D", SS,
	     "Indicates study has complete info (DICOM/genesis)", VM_1);

	// Group 0025 Elements
	_add("00250000", UL, "Group Length", VM_1);
	_add("00250010", LO, "GEHC Private Creator ID", VM_1);
	_add("00251006", SS, "Last pulse sequence used", VM_1);
	_add("00251007", SL, "Images in Series", VM_1);
	_add("00251010", SL, "Landmark Counter", VM_1);
	_add("00251011", SS, "Number of Acquisitions", VM_1);
	_add("00251014", SL, "Indicates no. of updates to header", VM_1);
	_add("00251017", SL, "Series Complete Flag", VM_1);
	_add("00251018", SL, "Number of images archived", VM_1);
	_add("00251019", SL, "Last image number used", VM_1);
	_add("0025101A", SH, "Primary Receiver Suite and Host", VM_1);
	_add("0025101B", OB, "Protocol Data Block (compressed)", VM_1);

	// Group 0027 Elements
	_add("00270000", UL, "Group Length", VM_1);
	_add("00270010", LO, "GEHC Private Creator ID", VM_1);
	_add("00271006", SL, "Image archive flag", VM_1);
	_add("00271010", SS, "Scout Type", VM_1);
	_add("00271030", SH, "Foreign Image Revision", VM_1);
	_add("00271031", SS, "Imaging Mode", VM_1);
	_add("00271032", SS, "Pulse Sequence", VM_1);
	_add("00271033", SL, "Imaging Options", VM_1);
	_add("00271035", SS, "Plane Type", VM_1);
	_add("00271036", SL, "Oblique Plane", VM_1);
	_add("00271040", SH, "RAS letter of image location", VM_1);
	_add("00271041", FL, "Image location", VM_1);
	_add("00271042", FL, "Center R coord of plane image", VM_1);
	_add("00271043", FL, "Center A coord of plane image", VM_1);
	_add("00271044", FL, "Center S coord of plane image", VM_1);
	_add("00271045", FL, "Normal R coord", VM_1);
	_add("00271046", FL, "Normal A coord", VM_1);
	_add("00271047", FL, "Normal S coord", VM_1);
	_add("00271048", FL, "R Coord of Top Right Corner", VM_1);
	_add("00271049", FL, "A Coord of Top Right Corner", VM_1);
	_add("0027104A", FL, "S Coord of Top Right Corner", VM_1);
	_add("0027104B", FL, "R Coord of Bottom Right Corner", VM_1);
	_add("0027104C", FL, "A Coord of Bottom Right Corner", VM_1);
	_add("0027104D", FL, "S Coord of Bottom Right Corner", VM_1);
	_add("00271060", FL, "Image dimension - X", VM_1);
	_add("00271061", FL, "Image dimension - Y", VM_1);
	_add("00271062", FL, "Number of Excitations", VM_1);

	// Group 0029 Elements
	_add("00290000", UL, "Group Length", VM_1);
	_add("00290010", SL, "GEHC Private Creator ID", VM_1);
	_add("00291015", SL, "Lower range of Pixels1", VM_1);
	_add("00291016", SL, "Upper range of Pixels1", VM_1);
	_add("00291017", SL, "Lower range of Pixels2", VM_1);
	_add("00291018", SL, "Upper range of Pixels2", VM_1);
	_add("0029101A", SL, "Len of tot hdr in bytes", VM_1);
	_add("00291026", SS, "Version of the hdr struct", VM_1);
	_add("00291034", SL, "Advantage comp. Overflow", VM_1);
	_add("00291035", SL, "Advantage comp. Underflow", VM_1);

	// Group 0043 Elements
	_add("00430000", UL, "Group Length", VM_1);
	_add("00430010", LO, "GEHC Private Creator ID", VM_1);
	_add("00431001", SS, "Bitmap of prescan options", VM_1);
	_add("00431002", SS, "Gradient offset in X", VM_1);
	_add("00431003", SS, "Gradient offset in Y", VM_1);
	_add("00431004", SS, "Gradient offset in Z", VM_1);
	_add("00431005", SS, "Img is original or unoriginal", VM_1);
	_add("00431006", SS, "Number of EPI shots", VM_1);
	_add("00431007", SS, "Views per segment", VM_1);
	_add("00431008", SS, "Respiratory rate, bpm", VM_1);
	_add("00431009", SS, "Respiratory trigger point", VM_1);
	_add("0043100A", SS, "Type of receiver used", VM_1);
	_add("0043100B", DS, "DB/dt Peak rate of change of gradient field",
	     VM_1);
	_add("0043100C", DS, "dB/dt Limits in units of percent", VM_1);
	_add("0043100D", DS, "PSD estimated limit", VM_1);
	_add("0043100E", DS, "PSD estimated limit in tesla per second", VM_1);
	_add("0043100F", DS, "Saravghead", VM_1);
	_add("00431010", US, "Window value", VM_1);
	_add("0043101C", SS, "GE image integrity", VM_1);
	_add("0043101D", SS, "Level value", VM_1);
	_add("00431028", OB, "Unique image iden", VM_1);
	_add("00431029", OB, "Histogram tables", VM_1);
	_add("0043102A", OB, "User defined data", VM_1);
	_add("0043102B", SS, "Private Scan Options", "4");
	_add("0043102C", SS, "Effective echo spacing", VM_1);
	_add("0043102D", SH,
	     "Filter Mode (String slop field 1 in legacy GE MR images)", VM_1);
	_add("0043102E", SH, "String slop field 2", VM_1);
	_add("0043102F", SS, "Image Type (real, imaginary, phase, magnitude)",
	     VM_1);
	_add("00431030", SS, "Vas collapse flag", VM_1);
	_add("00431031", DS, "RA cord of target recon center", "2");
	_add("00431032", SS, "Vas flags", VM_1);
	_add("00431033", FL, "Neg_scanspacing", VM_1);
	_add("00431034", IS, "Offset Frequency", VM_1);
	_add("00431035", UL, "User_usage_tag", VM_1);
	_add("00431036", UL, "User_fill_map_MSW", VM_1);
	_add("00431037", UL, "User_fill_map_LSW", VM_1);
	_add("00431038", FL,
	     "User data 25...User data 48 {User48=Effective Resolution for " +
	     "spiral}", "24");
	_add("00431039", IS,
	     "Slop_int_6... slop_int_9\n 6: b_value\n 7: private imaging " +
	     "options 2\n 8: ihtagging\n 9: ihtagspc", "4");
	_add("00431060", IS,
	     "Slop_int_10...slop_int_17\n 10: ihfcineim\n 11: ihfcinent\n 12: "+
	     "Reserved\n 13: oprtarr\n 14: averages\n 15: Current Station " +
	     "#\n 16: Total # of Stations\n 17: private imaging options 3",
	     "8");
	_add("00431061", UI, "Scanner Study Entity UID", VM_1);
	_add("00431062", SH,
	     "Scanner Table Entry (single gradient coil systems only)", VM_1);
	_add("0043106F", DS, "Scanner Table Entry + Gradient Coil Selected " +
	     "1:table entry, 2: table swing angle, 3: table lateral offset, " +
	     "4: gradient coil selected (multiple gradient systems only)",
	     "3-4");
	_add("00431070", LO, "Paradigm Name", VM_1);
	_add("00431071", ST, "Paradigm Description", VM_1);
	_add("00431072", UI, "Paradigm UID", VM_1);
	_add("00431073", US, "Experiment Type", VM_1);
	_add("00431074", US, "#rest volumes", VM_1);
	_add("00431075", US, "#active volumes", VM_1);
	_add("00431076", US, "#dummy scans", VM_1);
	_add("00431077", SH, "Application Name", VM_1);
	_add("00431078", SH, "Application Version", VM_1);
	_add("00431079", US, "Slices Per Volume", VM_1);
	_add("0043107A", US, "Expected Time Points", VM_1);
	_add("0043107B", FL, "Regressor Values", VM_1_N);
	_add("0043107C", FL, "Delay after slice group", VM_1);
	_add("0043107D", US, "Recon mode flag word", VM_1);
	_add("0043107E", LO, "PACC specific information", VM_1_N);
	_add("0043107F", DS, "Reserved", VM_1_N);
	_add("00431080", LO, "Coil ID Data", VM_1);
	_add("00431081", LO, "GE Coil Name", VM_1);
	_add("00431082", LO, "System Configuration Information", VM_1_N);
	_add("00431083", DS, "Asset R Factors", "1-2");
	_add("00431084", LO, "Additional Asset Data", "5");
	_add("00431085", UT, "Debug Data (text format)", VM_1);
	_add("00431086", OB, "Debug Data (binary format)", VM_1);
	_add("00431087", UT, "Reserved", VM_1);
	_add("00431088", UI, "PURE Acquisition Calibration Series UID", VM_1);
	_add("00431089", LO, "Governing Body, dB/dt, and SAR definition",
	     "3");
	_add("0043108A", CS, "Private In-Plane Phase Encoding Direction",
	     VM_1);
	_add("0043108B", OB, "FMRI Binary Data Block", VM_1);
	_add("0043108C", DS, "Voxel Location", "6");
	_add("0043108D", DS, "SAT Band Locations", VM_1_N);
	_add("0043108E", DS, "Spectro Prescan Values", "3");
	_add("0043108F", DS, "Spectro Parameters", "3");
	_add("00431090", LO, "SAR Definition", VM_1_N);
	_add("00431091", DS, "SAR value", VM_1_N);
	_add("00431092", LO, "Image Error Text", VM_1);
	_add("00431093", DS, "Spectro Quantitation Values", VM_1);
	_add("00431094", DS, "Spectro Ratio Values", VM_1);
	_add("00431095", LO, "Prescan Reuse String", VM_1);
	_add("00431096", CS, "Content Qualification", VM_1);
	_add("00431097", LO, "Image Filtering Parameters", "8");
	_add("00431098", UI, "ASSET Acquisition Calibration Series UID", VM_1);
	_add("00431099", LO, "Extended Options", VM_1_N);
	_add("0043109A", IS, "Rx Stack Identification", VM_1);
    }
}
