/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom;

/**
 * Data Dictionary of private DICOM Data Elements for Philips scanners.
 *
 * @version 23 February 2011
 */
public class PhilipsDataDictionary extends DataDictionary
{
    /** Single instance of the Philips Data Dictionary. */
    private static DataDictionary _dictionary = new PhilipsDataDictionary();

    /** Constructs a Philips Data Dictionary. */
    private PhilipsDataDictionary()
    {
    }

    /**
     * Gets the instance of the Philips Data Dictionary.
     *
     * @return Single instance of the Philips Data Dictionary.
     */
    public static DataDictionary getInstance()
    {
	return _dictionary;
    }

    /**
     * Determines if this Data Dictionary has private Data Elements from the
     * manufacturer.
     *
     * @param manufacturer Manufacturer.
     *
     * @return True if there are private Data Elements; false otherwise.
     */
    public static boolean isDictionaryFor(String manufacturer)
    {
	return "Philips Medical Systems".equals(manufacturer.trim());
    }

    /** Loads the registry of DICOM Data Elements. */
    protected void _loadRegistry()
    {
	// Group 2001 Elements
	_add("20010000", UL, "Group Length", VM_1);
	_add("20010010", LO, "Private Creator Group 2001", VM_1);
	_add("20011001", FL, "Chemical Shift", VM_1);
	_add("20011002", IS, "Chemical Shift Number MR", VM_1);
	_add("20011003", FL, "Diffusion b-value", VM_1);
	_add("20011004", CS, "Diffusion Direction", VM_1);
	_add("20011006", CS, "Image Enhanced", VM_1);
	_add("20011007", CS, "Image Type ED ES", VM_1);
	_add("20011008", IS, "Phase Number", VM_1);
	_add("20011009", FL, "", VM_1);
	_add("2001100A", IS, "Slice Number MR", VM_1);
	_add("2001100B", CS, "", VM_1);
	_add("2001100C", CS, "", VM_1);
	_add("2001100E", CS, "", VM_1);
	_add("2001100F", SS, "", VM_1);
	_add("20011010", CS, "", VM_1);
	_add("20011011", FL, "Diffusion Echo Time", VM_1);
	_add("20011012", CS, "Dynamic Series", VM_1);
	_add("20011013", SL, "EPI Factor", VM_1);
	_add("20011014", SL, "Number of Echoes", VM_1);
	_add("20011015", SS, "Number of Locations", VM_1);
	_add("20011016", SS, "Number of PC Directions", VM_1);
	_add("20011017", SL, "Number of Phases MR", VM_1);
	_add("20011018", SL, "Number of Slices MR", VM_1);
	_add("20011019", CS, "Partial Matrix Scanned", VM_1);
	_add("2001101A", FL, "PC Velocity", VM_1_N);
	_add("2001101B", FL, "Prepulse Delay", VM_1);
	_add("2001101C", CS, "Prepulse Type", VM_1);
	_add("2001101D", IS, "Reconstruction Number MR", VM_1);
	_add("2001101F", CS, "Respiration Sync", VM_1);
	_add("20011020", LO, "Scanning Technique Description MR", VM_1);
	_add("20011021", CS, "SPIR", VM_1);
	_add("20011022", FL, "Water Fat Shift", VM_1);
	_add("20011023", DS, "Flip Angle Philips", VM_1);
	_add("20011024", CS, "", VM_1);
	_add("20011025", SH, "Echo Time Display MR", VM_1);
	_add("2001102D", SS, "Stack Number Of Slices", VM_1);
	_add("20011032", FL, "Stack Radial Angle", VM_1);
	_add("20011033", CS, "Stack Radial Axis", VM_1);
	_add("20011035", SS, "Stack Slice Number", VM_1);
	_add("20011036", CS, "Stack Type", VM_1);
	_add("2001103F", CS, "Zoom Mode", VM_1);
	_add("2001105F", SQ, "Stack Sequence", VM_1);
	_add("20011060", SL, "Number of Stacks", VM_1);
	_add("20011061", CS, "", VM_1);
	_add("20011062", CS, "", VM_1);
	_add("20011063", CS, "Examination Source", VM_1);
	_add("2001107B", IS, "Acquisition Number", VM_1);
	_add("20011081", IS, "Number of Dynamic Scans", VM_1);
	_add("20011082", IS, "", VM_1);
	_add("20011083", DS, "", VM_1);
	_add("20011084", DS, "", VM_1);
	_add("20011085", DS, "", VM_1);
	_add("20011086", IS, "", VM_1);
	_add("20011087", SH, "", VM_1);
	_add("20011088", DS, "", VM_1);
	_add("20011089", DS, "", VM_1);
	_add("2001108A", DS, "", VM_1);
	_add("2001108B", SH, "", VM_1);
	_add("200110F1", FL, "Prospective Motion Correction", VM_1_N);
	_add("200110F2", FL, "Retrospective Motion Correction", VM_1);

	// Group 2005 Elements
	_add("20050000", UL, "Group Length", VM_1);
	_add("20050010", LO, "Private Creator Group 2005", VM_1);
	_add("20050011", LO, "", VM_1);
	_add("20050012", LO, "", VM_1);
	_add("20050013", LO, "", VM_1);
	_add("20051000", FL, "", VM_1);
	_add("20051001", FL, "", VM_1);
	_add("20051002", FL, "", VM_1);
	_add("20051004", CS, "", VM_1);
	_add("20051008", FL, "", VM_1);
	_add("20051009", FL, "", VM_1);
	_add("2005100A", FL, "", VM_1);
	_add("2005100B", FL, "", VM_1);
	_add("2005100C", FL, "", VM_1);
	_add("2005100D", FL, "", VM_1);
	_add("2005100E", FL, "", VM_1);
	_add("2005100F", DS, "", VM_1);
	_add("20051010", DS, "", VM_1);
	_add("20051011", CS, "", VM_1);
	_add("20051012", CS, "", VM_1);
	_add("20051013", CS, "", VM_1);
	_add("20051014", CS, "", VM_1);
	_add("20051015", CS, "", VM_1);
	_add("20051016", CS, "", VM_1);
	_add("20051017", CS, "", VM_1);
	_add("20051019", CS, "", VM_1);
	_add("2005101A", SS, "", VM_1);
	_add("2005101B", CS, "", VM_1);
	_add("2005101C", CS, "", VM_1);
	_add("2005101D", SS, "", VM_1);
	_add("20051020", SL, "Number of Chemical Shift", VM_1);
	_add("20051021", SS, "", VM_1);
	_add("20051022", IS, "", VM_1);
	_add("20051023", SS, "", VM_1);
	_add("20051025", SS, "", VM_1);
	_add("20051026", CS, "", VM_1);
	_add("20051027", CS, "", VM_1);
	_add("20051028", CS, "", VM_1);
	_add("20051029", CS, "", VM_1);
	_add("2005102A", IS, "", VM_1);
	_add("2005102B", SS, "", VM_1);
	_add("2005102C", CS, "", VM_1);
	_add("2005102D", IS, "", VM_1);
	_add("2005102E", CS, "", VM_1);
	_add("2005102F", CS, "", VM_1);
	_add("20051030", FL, "", VM_1);
	_add("20051031", CS, "", VM_1);
	_add("20051033", FL, "", VM_1);
	_add("20051034", CS, "", VM_1);
	_add("20051035", CS, "", VM_1);
	_add("20051036", CS, "", VM_1);
	_add("20051037", CS, "", VM_1);
	_add("20051038", CS, "", VM_1);
	_add("20051039", CS, "", VM_1);
	_add("2005103B", CS, "", VM_1);
	_add("2005103C", CS, "", VM_1);
	_add("2005103D", SS, "", VM_1);
	_add("2005103E", SL, "", VM_1_N);
	_add("2005105F", CS, "", VM_1);
	_add("20051060", IS, "", VM_1);
	_add("20051061", CS, "", VM_1);
	_add("20051063", SS, "", VM_1);
	_add("2005106E", CS, "", VM_1);
	_add("2005106F", CS, "", VM_1);
	_add("20051071", FL, "", VM_1);
	_add("20051072", FL, "", VM_1);
	_add("20051073", FL, "", VM_1);
	_add("20051074", FL, "", VM_1);
	_add("20051075", FL, "", VM_1);
	_add("20051076", FL, "", VM_1);
	_add("20051078", FL, "", VM_1);
	_add("20051079", FL, "", VM_1);
	_add("2005107A", FL, "", VM_1);
	_add("2005107B", CS, "", VM_1);
	_add("2005107E", FL, "", VM_1);
	_add("20051080", SQ, "", VM_1);
	_add("20051081", CS, "", VM_1);
	_add("20051083", SQ, "", VM_1);
	_add("20051084", SQ, "", VM_1);
	_add("20051085", SQ, "", VM_1);
	_add("20051086", SS, "", VM_1);
	_add("2005109E", SQ, "", VM_1);
	_add("200510A0", FL, "", VM_1);
	_add("200510A1", CS, "Syncra Scan Type", VM_1);
	_add("200510A2", CS, "", VM_1);
	_add("200510A3", IS, "", VM_1);
	_add("200510A4", IS, "", VM_1);
	_add("200510A5", IS, "", VM_1);
	_add("200510A6", IS, "", VM_1);
	_add("200510A7", CS, "", VM_1);
	_add("200510A8", DS, "", VM_1);
	_add("200510A9", CS, "", VM_1);
	_add("200510B0", FL, "Diffusion Direction RL", VM_1);
	_add("200510B1", FL, "Diffusion Direction AP", VM_1);
	_add("200510B2", FL, "Diffusion Direction FH", VM_1);
	_add("200510C0", CS, "", VM_1);
	_add("20051199", UL, "", VM_1);
	_add("20051200", UL, "", VM_1);
	_add("20051201", UL, "", VM_1);
	_add("20051213", UL, "", VM_1);
	_add("20051245", SS, "", VM_1);
	_add("20051249", SS, "", VM_1);
	_add("20051251", SS, "", VM_1);
	_add("20051252", SS, "", VM_1);
	_add("20051253", SS, "", VM_1);
	_add("20051325", CS, "", VM_1);
	_add("20051326", FL, "", VM_1);
	_add("20051327", CS, "", VM_1);
	_add("20051328", CS, "", VM_1);
	_add("20051329", FL, "", VM_1);
	_add("20051330", CS, "", VM_1_N);
	_add("20051331", SS, "", VM_1);
	_add("20051333", FL, "", VM_1_N);
	_add("20051334", CS, "", VM_1);
	_add("20051335", CS, "", VM_1);
	_add("20051336", FL, "", VM_1);
	_add("20051337", FL, "", VM_1);
	_add("20051338", FL, "", VM_1);
	_add("20051339", SS, "", VM_1_N);
	_add("20051390", CS, "", VM_1);
	_add("20051340", CS, "", VM_1);
	_add("20051341", CS, "", VM_1);
	_add("20051342", CS, "", VM_1);
	_add("20051343", CS, "", VM_1);
	_add("20051344", SS, "", VM_1_N);
	_add("20051345", CS, "", VM_1);
	_add("20051346", CS, "", VM_1);
	_add("20051347", FL, "", VM_1);
	_add("20051348", CS, "", VM_1);
	_add("20051349", FL, "", VM_1);
	_add("20051350", FL, "", VM_1_N);
	_add("20051351", SS, "", VM_1);
	_add("20051352", SS, "", VM_1);
	_add("20051355", FL, "", VM_1_N);
	_add("20051356", CS, "", VM_1);
	_add("20051357", SS, "", VM_1);
	_add("20051359", FL, "", VM_1);
	_add("20051360", FL, "", VM_1);
	_add("20051361", FL, "", VM_1_N);
	_add("20051362", FL, "", VM_1);
	_add("20051363", FL, "", VM_1);
	_add("20051364", CS, "", VM_1);
	_add("20051370", SS, "", VM_1);
	_add("20051371", SQ, "", VM_1);
	_add("20051381", IS, "", VM_1);
	_add("20051383", UL, "", VM_1);
	_add("20051392", IS, "", VM_1);
	_add("20051393", IS, "", VM_1);
	_add("20051396", CS, "", VM_1);
	_add("20051398", CS, "", VM_1);
	_add("20051399", CS, "", VM_1);
	_add("20051400", CS, "", VM_1);
	_add("20051401", UL, "", VM_1);
	_add("20051403", UL, "", VM_1);
	_add("20051406", SS, "", VM_1);
	_add("20051409", DS, "", VM_1);
	_add("2005140A", DS, "", VM_1);
	_add("2005140E", SQ, "", VM_1);
	_add("2005140F", SQ, "", VM_1);
	_add("20051412", IS, "", VM_1);
	_add("20051413", IS, "", VM_1);
	_add("20051414", SL, "", VM_1);
	_add("20051415", SL, "", VM_1);
	_add("20051416", CS, "", VM_1);
	_add("20051418", CS, "", VM_1_N);
	_add("20051419", CS, "", VM_1_N);
	_add("2005141A", CS, "", VM_1);
	_add("2005141B", IS, "", VM_1);
	_add("2005141C", IS, "", VM_1);
	_add("2005141D", IS, "", VM_1);
	_add("20051426", CS, "", VM_1);
	_add("20051428", SL, "", VM_1);
	_add("2005142A", CS, "", VM_1);
	_add("2005142B", CS, "", VM_1);
	_add("2005142C", CS, "", VM_1);
	_add("2005142D", CS, "", VM_1);
	_add("20051432", CS, "", VM_1);
	_add("20051435", CS, "", VM_1);
	_add("2005143B", CS, "", VM_1);
	_add("2005143C", FL, "", VM_1);
	_add("2005143D", FL, "", VM_1);
	_add("2005143E", FL, "", VM_1);
	_add("2005143F", CS, "", VM_1);
	_add("20051440", FL, "", VM_1);
	_add("20051441", FL, "", VM_1);
	_add("20051442", FL, "", VM_1);
	_add("20051443", FL, "", VM_1);
	_add("20051444", IS, "", VM_1);
	_add("20051445", CS, "", VM_1);
	_add("20051447", FL, "", VM_1);
	_add("2005144A", FL, "", VM_1);
	_add("2005144B", FL, "", VM_1);
	_add("2005144C", FL, "", VM_1);
	_add("2005144D", CS, "", VM_1);
	_add("2005144E", CS, "", VM_1);
	_add("2005144F", CS, "", VM_1);
	_add("20051450", SS, "", VM_1);
    }
}
