/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.decoder;

import edu.ucla.loni.jdicom.DataSet;
import java.io.IOException;
import java.io.InputStream;
import javax.imageio.stream.ImageInputStream;

/**
 * Reader for reading and decoding a DICOM set of Data Elements.
 *
 * @version 11 March 2011
 */
public class DataSetReader
{
    /** Stream to read the Data Set from. */
    private ImageInputStream _stream;

    /** Offset in the stream to the start of the encoded Data Set. */
    private long _offset;

    /** Decoding parameters. */
    private DecodeParams _params;

    /** Length (in bytes) of the encoded DataSet. */
    private int _dataSetLength;

    /** 8 digit hexadecimal number marking the end of the encoded DataSet. */
    private String _dataElementDelimiterTag;

    /**
     * Constructs a Data Set Reader that returns an empty Data Set.
     *
     * @param stream Stream to read the Data Set from.
     * @param params Decoding parameters.
     */
    public DataSetReader()
    {
	_dataSetLength = -1;
    }

    /**
     * Constructs a Data Set Reader.
     *
     * @param stream Stream to read the Data Set from.
     * @param params Decoding parameters.
     * @param dataSetLength Length (in bytes) of the encoded DataSet.
     *
     * @throws IOException If an I/O error occurs.
     */
    public DataSetReader(ImageInputStream stream, DecodeParams params,
			 int dataSetLength) throws IOException
    {
	_stream = stream;
	_offset = stream.getStreamPosition();
	_params = params;
	_dataSetLength = dataSetLength;
    }

    /**
     * Constructs a Data Set Reader.
     *
     * @param stream Stream to read the Data Set from.
     * @param params Decoding parameters.
     * @param dataElementDelimiterTag 8 digit hexadecimal number marking the
     *                                end of the encoded DataSet.
     *
     * @throws IOException If an I/O error occurs.
     */
    public DataSetReader(ImageInputStream stream, DecodeParams params,
			 String dataElementDelimiterTag) throws IOException
    {
	this(stream, params, -1);

	_dataElementDelimiterTag = dataElementDelimiterTag;
    }

    /**
     * Gets the Data Set.
     *
     * @return Data Set.
     *
     * @throws IOException If an I/O error occurs.
     */
    public DataSet getDataSet() throws IOException
    {
	// No Data Set
	if (_dataSetLength == -1 && _dataElementDelimiterTag == null) {
	    return new DataSet();
	}

	// Reset the stream location to the beginning of the Data Set
	_stream.seek(_offset);

	// Convert to an input stream
	InputStream inputStream = new InputStreamAdapter(_stream);
	MeasuredInputStream stream = new MeasuredInputStream(inputStream);

	// Data Set length is specified
	if (_dataSetLength > -1) {
	    DataSetDecoder decoder = new DataSetDecoder(stream, _params,
							_dataSetLength);
	    return decoder.getDataSet();
	}

	// Delimiter tag is specified
	DataSetDecoder decoder = new DataSetDecoder(stream, _params,
						    _dataElementDelimiterTag);
	return decoder.getDataSet();
    }

    /** 
     * Gets the contents of the Data Set Reader.
     *
     * @return String representation of the contents of the Data Set Reader.
     */
    public String toString()
    {
	return "DATA SET READER [offset = " + _offset + "]";
    }
}
