/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.node;

import edu.ucla.loni.jdicom.DataElement;
import edu.ucla.loni.jdicom.DataSet;
import edu.ucla.loni.jdicom.plugin.DicomMetadataFormat;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;

/**
 * DICOM List Node that represents a Data Set.
 *
 * @version 9 April 2010
 */
public class DataSetNode extends DicomListNode
{
    /** Mapping from Data Element names to Data Elements. */
    private HashMap _tagsToElements;

    /**
     * Constructs a Data Set Node.
     *
     * @param dataSet Data Set to convert.
     */
    public DataSetNode(DataSet dataSet)
    {
	this( _getChildren(dataSet) );
    }

    /**
     * Constructs a Data Set Node.
     *
     * @param children Child Nodes from the Data Set.
     */
    private DataSetNode(ArrayList children)
    {
	super(children);

	_tagsToElements = new HashMap(children.size());

	// Populate the hash map
	Iterator iter = children.iterator();
	while ( iter.hasNext() ) {
	    DataElementNode dataElement = (DataElementNode)iter.next();
	    _tagsToElements.put(dataElement.getTag(), dataElement);
	}
    }

    /**
     * Gets the first value of the Data Element.
     *
     * @param tag Tag of the Data Element.
     * @param defaultValue Default value.
     *
     * @return First value of the Data Element or the default value if no
     *         values exist.
     */
    public String getDataElementValue(String tag, String defaultValue)
    {
	DataElementNode dataElement = getDataElement(tag);

	// No Data Element
	if (dataElement == null) { return defaultValue; }

	// Get the first value
	String value = dataElement.getValue();

	// No values
	if (value == null) { return defaultValue; }

	// Return the value
	return value;
    }

    /**
     * Gets the first value of the Data Element.
     *
     * @param tag Tag of the Data Element.
     * @param defaultValue Default value.
     *
     * @return First value of the Data Element or the default value if no
     *         values exist.
     */
    public int getDataElementValue(String tag, int defaultValue)
    {
	DataElementNode dataElement = getDataElement(tag);

	// No Data Element
	if (dataElement == null) { return defaultValue; }

	// Get the first value
	String value = dataElement.getValue();

	// No values
	if (value == null) { return defaultValue; }

	// Convert to an integer
	try { return Integer.parseInt(value); }
	catch (Exception e) { return defaultValue; }
    }

    /**
     * Gets the Data Element.
     *
     * @param tag Tag of the Data Element.
     *
     * @return Data Element with the tag or null if none is in the Data Set.
     */
    public DataElementNode getDataElement(String tag)
    {
	return (DataElementNode)_tagsToElements.get(tag);
    }

    /**
     * Gets the Data Elements.
     *
     * @return Data Elements.
     */
    public Iterator getDataElements()
    {
	return new DataElementNodeIterator(getChildNodeIterator(),
					   _tagsToElements);
    }

    /**
     * Adds the Data Element to the Data Set.
     *
     * @param tag Tag of the Data Element.
     * @param vr Value representation of the Data Element.
     *
     * @return Data Element that was added.
     */
    public DataElementNode addDataElement(String tag, String vr)
    {
	try {
	    DataElementNode dataElement = new DataElementNode(tag, vr);
	    return (DataElementNode)appendChild(dataElement);
	}
	catch (DOMException e) {
	    String msg = "Unable to add Data Element:  " + tag + "(" + vr + ")";
	    throw new RuntimeException(msg, e);
	}
    }

    /**
     * Gets the Node name.
     *
     * @return Name of the Node.
     */
    public String getNodeName()
    {
	return DicomMetadataFormat.DATA_SET_NAME;
    }

    /**
     * Inserts the Node before the existing child Node.  If the existing child
     * Node is null, the new Node is inserted at the end of the list of
     * children.
     *
     * @param newChild Node to insert.
     * @param refChild Node before which the new Node is inserted.
     *
     * @return Node that is inserted.
     *
     * @throws DOMException If the child is not a Data Element Node.
     * @throws DOMException If the reference child is not a child of this Node.
     */
    public Node insertBefore(Node newChild, Node refChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof DataElementNode) ) {
	    String msg = "Illegal new child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DataElementNode newNode = (DataElementNode)newChild;

	// Insert the child
	Node node = super.insertBefore(newChild, refChild);

	// Add to the hash map
	_tagsToElements.put(newNode.getTag(), newNode);

	// Return the Node that is inserted
	return node;
    }

    /**
     * Replaces the child Node with the given child Node.
     *
     * @param newChild Node to insert.
     * @param oldChild Node being replaced.
     *
     * @return Node that is replaced.
     *
     * @throws DOMException If a child is not a Data Element Node.
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node replaceChild(Node newChild, Node oldChild) throws DOMException
    {
	// Illegal child
	if ( !(newChild instanceof DataElementNode) ||
	     !(oldChild instanceof DataElementNode) )
	    {
		String msg = "Illegal child node";
		throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	    }
	DataElementNode newNode = (DataElementNode)newChild;
	DataElementNode oldNode = (DataElementNode)oldChild;

	// Replace the child
	Node node = super.replaceChild(newChild, oldChild);

	// Adjust the hash map
	_tagsToElements.put(newNode.getTag(), newNode);
	_tagsToElements.remove(oldNode.getTag());

	// Return the Node that is replaced
	return node;
    }

    /**
     * Removes the child Node from the list of children Nodes.
     *
     * @param oldChild Node being removed.
     *
     * @return Node that is removed.
     *
     * @throws DOMException If the child is not a Data Element Node.
     * @throws DOMException If the old child is not a child of this Node.
     */
    public Node removeChild(Node oldChild) throws DOMException
    {
	// Illegal old child
	if ( !(oldChild instanceof DataElementNode) ) {
	    String msg = "Illegal old child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DataElementNode oldNode = (DataElementNode)oldChild;

	// Remove the child
	Node node = super.removeChild(oldChild);

	// Remove from the hash map
	_tagsToElements.remove(oldNode.getTag());

	// Return the Node that is removed
	return node;
    }

    /**
     * Adds the Node to the end of the list of children Nodes.
     *
     * @param newChild Node to add.
     *
     * @return Node that is added.
     *
     * @throws DOMException If the child is not a Data Element Node.
     */
    public Node appendChild(Node newChild) throws DOMException
    {
	// Illegal new child
	if ( !(newChild instanceof DataElementNode) ) {
	    String msg = "Illegal new child node";
	    throw new DOMException(DOMException.WRONG_DOCUMENT_ERR, msg);
	}
	DataElementNode newNode = (DataElementNode)newChild;

	// Add the child
	Node node = super.appendChild(newChild);

	// Add to the hash map
	_tagsToElements.put(newNode.getTag(), newNode);

	// Return the Node that is added
	return node;
    }

    /**
     * Gets a duplicate of this Node.  The duplicate Node has no parent.  If a
     * shallow clone is being performed, the duplicate Node will not have any
     * children or siblings.  If a deep clone is performed, the duplicate Node
     * will form the root of a completely cloned subtree.
     *
     * @param deep True for a recursive clone, false for cloning only the Node
     *             itself.
     *
     * @return Duplicate Node.
     */
    public Node cloneNode(boolean deep)
    {
	ArrayList children = new ArrayList(0);
	if (deep) {
	    children = new ArrayList( getLength() );

	    for (int i = 0; i < getLength(); i++) {
		Node child = item(i);
		children.add( child.cloneNode(true) );
	    }
	}

	return new DataSetNode(children);
    }

    /**
     * Creates child Nodes from the Data Set.
     *
     * @param dataSet Data Set to represent.
     *
     * @return Child Nodes from the Data Set.
     */
    private static ArrayList _getChildren(DataSet dataSet)
    {
	// Initialize the list
	ArrayList children = new ArrayList(dataSet.getNumberOfDataElements());

	// Add the Data Elements
	Enumeration enumElems = dataSet.getDataElements();
	while ( enumElems.hasMoreElements() ) {
	    DataElement dataElement = (DataElement)enumElems.nextElement();
	    children.add( new DataElementNode(dataElement) );
	}

	return children;
    }
}
