/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.write;

import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.io.IOException;
import java.nio.ByteOrder;
import javax.imageio.stream.ImageOutputStream;

/**
 * DICOM block for image pixel data.
 *
 * @version 15 March 2011
 */
public class ImageBlock extends DicomBlock
{
    /**
     * Constructs an Image Block.
     *
     * @param outputStream Output stream where the block is written.
     * @param transferSyntax Transfer syntax that determines the encoding.
     * @param dataBuffer Data Buffer containing the pixel data.
     *
     * @throws IOException If an error occurs during writing.
     */
    public ImageBlock(ImageOutputStream outputStream, String transferSyntax,
		      DataBuffer dataBuffer)
	throws IOException
    {
	super(outputStream, transferSyntax);

	// Write an 8 bit grayscale image
	if (dataBuffer instanceof DataBufferByte) {
	    outputStream.write( ((DataBufferByte)dataBuffer).getData() );
	}

	// Write a 16 bit grayscale image
	else if (dataBuffer instanceof DataBufferUShort) {
	    short[] data = ((DataBufferUShort)dataBuffer).getData();

	    // Write the image, byte swapping if necessary
	    if (_byteSwap()) {
		outputStream.setByteOrder(ByteOrder.LITTLE_ENDIAN);
	    }
	    outputStream.writeShorts(data, 0, data.length);
	    outputStream.setByteOrder(ByteOrder.BIG_ENDIAN);
	}

	// Write a 32 bit RGB color image
	else if (dataBuffer instanceof DataBufferInt) {
	    int[] data = ((DataBufferInt)dataBuffer).getData();
	    outputStream.write( _separateRgbComponents(data) );
	}

	// Unrecognized type
	else {
	    String msg = "Unable to recognize the image type.";
	    throw new IOException(msg);
	}

	// Update the size
	_setSize();
    }

    /**
     * Separates the RGB components of each integer pixel into 3 bytes.
     *
     * @param intArray Array of integers where each element represents an ARGB
     *                 pixel.
     *
     * @return Byte array where each consecutive 3 bytes represents the RGB
     *         components of an ARGB pixel in the integer array.
     */
    private byte[] _separateRgbComponents(int[] intArray)
    {
	// Create a byte array to hold the RGB components
	byte[] byteArray = new byte[3*intArray.length];

	// Separate the RGB components of each integer pixel
	for (int i = 0; i < intArray.length; i++) {
	    int pixel = intArray[i];

	    byteArray[3*i]   = (byte)( (pixel >> 16) & 0x000000ff );
	    byteArray[3*i+1] = (byte)( (pixel >>  8) & 0x000000ff );
	    byteArray[3*i+2] = (byte)( (pixel      ) & 0x000000ff );
	}

	// Return the RGB components
	return byteArray;
    }
}
