/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.jdicom.plugin.write;

import edu.ucla.loni.jdicom.DataElement;
import edu.ucla.loni.jdicom.encoder.ByteSwappedOutputStream;
import edu.ucla.loni.jdicom.encoder.DataElementEncoder;
import edu.ucla.loni.jdicom.plugin.OutputStreamAdapter;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.Vector;
import javax.imageio.stream.ImageOutputStream;

/**
 * DICOM block for a sequence of images.
 *
 * @version 17 March 2011
 */
public class ImageSequence extends DicomBlock
{
    /** Image blocks in the sequence. */
    private Vector _imageBlocks;

    /** Transfer syntax that determines the encoding. */
    private String _transferSyntax;

    /** Value Representation for the pixel data. */
    private String _vr;

    /**
     * Constructs an Image Sequence.
     *
     * @param outputStream Output stream where the block is written.
     * @param transferSyntax Transfer syntax that determines the encoding.
     *
     * @throws IOException If an error occurs during writing.
     */
    public ImageSequence(ImageOutputStream outputStream, String transferSyntax)
	throws IOException
    {
	super(outputStream, transferSyntax);

	_transferSyntax = transferSyntax;
	_vr = "OB";
	_imageBlocks = new Vector();
    }

    /**
     * Moves the block to another part of the output stream.
     *
     * @param newOffset New offset in the stream for the block.
     *
     * @throws IOException If an error occurs during reading or writing.
     */
    public void moveTo(long newOffset) throws IOException
    {
	// Determine the shift
	long shift = newOffset - getOffset();

	// Write a new header
	_writeHeader(newOffset);

	// Move each image block
	for (int i = _imageBlocks.size()-1; i >= 0; i--) {
	    ImageBlock imageBlock = (ImageBlock)_imageBlocks.get(i);
	    imageBlock.moveTo(imageBlock.getOffset() + shift);
	}

	// Set the new offset
	_setOffset(newOffset);
    }

    /**
     * Adds the image to the end of the block.
     *
     * @param dataBuffer Data Buffer containing the pixel data.
     * @param width Image width.
     * @param height Image height.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void append(DataBuffer dataBuffer, int width, int height)
	throws IOException
    {
	// First image
	if (_imageBlocks.size() == 0) {

	    // Determine the VR
	    _vr = "OB";
	    if (dataBuffer instanceof DataBufferByte) {
		int imageSize = width*height;

		// Byte data corresponds to more than 1 byte per pixel
		if (imageSize > 0 && dataBuffer.getSize()/imageSize > 1) {
		    _vr = "OW";
		}
	    }
	    else { _vr = "OW"; }

	    // Determine the number of image bytes written
	    int bytesPerPixel = 1;
	    if (dataBuffer instanceof DataBufferUShort) { bytesPerPixel = 2; }
	    if (dataBuffer instanceof DataBufferInt) { bytesPerPixel = 3; }
	    int pixelDataSize = bytesPerPixel*dataBuffer.getSize();

	    // Write the pixel data header at the beginning of the block
	    _writeHeader(getOffset(), pixelDataSize);

	    // Update the size
	    _setSize();
	}

	// Get the output stream at the end of the block
	ImageOutputStream outputStream = _getOutputStreamAtEnd();

	// Write a new image block
	ImageBlock imageBlock = new ImageBlock(outputStream, _transferSyntax,
					       dataBuffer);
	_imageBlocks.add(imageBlock);

	// Update the size
	_setSize();
    }

    /**
     * Finishes writing the block.
     *
     * @throws IOException If an error occurs during writing.
     */
    public void finish() throws IOException
    {
	// Update the pixel data header if more than 1 image
	if (_imageBlocks.size() > 1) {

	    // Overwrite the header at the beginning of the block
	    _writeHeader( getOffset() );
	}
    }

    /**
     * Writes the pixel data header at the offset.
     *
     * @param offset Offset to write the pixel data header at.
     *
     * @throws IOException If an error occurs during writing.
     */
    private void _writeHeader(long offset) throws IOException
    {
	int pixelDataSize = 0;

	// Compute the number of pixel data bytes
	Iterator iter = _imageBlocks.iterator();
	while ( iter.hasNext() ) {
	    ImageBlock imageBlock = (ImageBlock)iter.next();
	    pixelDataSize += imageBlock.getSize();
	}

	// Overwrite the header at the beginning of the block
	_writeHeader(offset, pixelDataSize);
    }

    /**
     * Writes the pixel data header at the offset.
     *
     * @param offset Offset to write the pixel data header at.
     * @param pixelDataSize Number of bytes of pixel data.
     *
     * @throws IOException If an error occurs during writing.
     */
    private void _writeHeader(long offset, int pixelDataSize) throws IOException
    {
	// Get the output stream at the offset
	ImageOutputStream outputStream = _getOutputStreamAtEnd();
	outputStream.seek(offset);

	// Create the encoder
	OutputStream os = new OutputStreamAdapter(outputStream);
	DataElementEncoder encoder = new DataElementEncoder(os,_transferSyntax);

	// Write the start of the pixel Data Element
	DataElement element = new DataElement("7FE00010", _vr, "Pixel Data");
	encoder.writeDataElement(element, false);

	// Write the value length
	ByteSwappedOutputStream bsos = new ByteSwappedOutputStream(os);
	bsos.writeInt(pixelDataSize, _byteSwap());
    }
}
