/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd;

import java.util.Iterator;
import java.util.LinkedHashSet;

/**
 * NRRD Header.
 *
 * @version 8 November 2010
 */
public class NrrdHeader
{
    /** Magic number. */
    private String _magicNumber;

    /** Elements of the NRRD header. */
    private LinkedHashSet _elements;

    /**
     * Constructs an NRRD Header.
     *
     * @param magicNumber Magic number.
     */
    public NrrdHeader(String magicNumber)
    {
	_magicNumber = magicNumber;
	_elements = new LinkedHashSet();
    }

    /**
     * Gets the magic number.
     *
     * @return Magic number.
     */
    public String getMagicNumber()
    {
	return _magicNumber;
    }

    /**
     * Gets the field descriptor.
     *
     * @param identifier Field identifier.
     *
     * @return Field descriptor or null if the field does not exist.
     */
    public String getFieldDescriptor(String identifier)
    {
	// Search for the Field
	Iterator iter = _elements.iterator();
	while ( iter.hasNext() ) {
	    NrrdHeaderElement next = (NrrdHeaderElement)iter.next();

	    if (next instanceof Field) {
		Field field = (Field)next;
		if ( field.getIdentifier().equals(identifier) ) {
		    return field.getDescriptor();
		}
	    }
	}

	// Field not found
	return null;
    }

    /**
     * Gets the header elements.
     *
     * @return Header elements.
     */
    public NrrdHeaderElement[] getElements()
    {
	return (NrrdHeaderElement[])_elements.toArray(new NrrdHeaderElement[0]);
    }

    /**
     * Gets a Key-Value Pair Set with the name.  A new Set is created if the
     * Set does not exist.
     *
     * @param name Name of the Set.
     *
     * @return Key-Value Pair Set with the name.
     */
    public KeyValuePairSet getPairSet(String name)
    {
	KeyValuePairSet newSet = new KeyValuePairSet(name);

	// Search for the Set
	Iterator iter = _elements.iterator();
	while ( iter.hasNext() ) {
	    NrrdHeaderElement next = (NrrdHeaderElement)iter.next();
	    if ( next.equals(newSet) ) { return (KeyValuePairSet)next; }
	}

	// Otherwise add a new Set
	addElement(newSet);
	return newSet;
    }

    /**
     * Adds a header element.
     *
     * @param element Element to add.
     */
    public void addElement(NrrdHeaderElement element)
    {
	// Remove the element if it already exists
	if ( _elements.contains(element) ) { _elements.remove(element); }

	// Add the element
	_elements.add(element);
    }

    /**
     * Gets a string representation of this Object.
     *
     * @return String representation of this Object.
     */
    public String toString()
    {
	StringBuffer buffer = new StringBuffer("MAGIC = " + _magicNumber +"\n");

	NrrdHeaderElement[] elements = getElements();
	for (int i = 0; i < elements.length; i++) {
	    buffer.append(elements[i] + "\n");
	}

	return buffer.toString();
    }
}
