/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd;

import javax.imageio.IIOException;
import javax.imageio.stream.ImageInputStream;

/**
 * Class that decodes an encoded NRRD header.
 *
 * @version 25 October 2010
 */
public class NrrdHeaderDecoder
{
    /** Decoded header. */
    private NrrdHeader _header;

    /**
     * Constructs an NRRD Header Decoder.
     *
     * @param inputStream Image Input Stream to read from.
     *
     * @throws IIOException If an I/O error occurs.
     */
    public NrrdHeaderDecoder(ImageInputStream inputStream) throws IIOException
    {
	try {

	    // Get the magic number
	    String line = inputStream.readLine();
	    _header = new NrrdHeader(line);

	    // Read lines until reach a zero character line or end-of-file
	    while (line != null && line.length() > 0) {

		// Ignore comments
		if ( line.trim().startsWith("#") ) {
		    line = inputStream.readLine();
		    continue;
		}

		// Key-value pair
		int delimIndex = line.indexOf(":=");
		if (delimIndex != -1) {

		    // Get the key
		    String key = line.substring(0, delimIndex);
		    key = key.trim();

		    // Get the value
		    String value = line.substring(delimIndex+2);
		    value = value.trim();

		    // Part of a key set:  subKey _ NNN := value
		    int index = key.lastIndexOf("_");
		    String nnn = key.substring(index+1);
		    if ( _isNumber(nnn) ) {

			// Get a Key-Value Pair Set
			String subKey = key.substring(0,index);
			KeyValuePairSet pairSet = _header.getPairSet(subKey);

			// Add to the Set
			pairSet.setPair(nnn, value);
		    }

		    // Otherwise add a key-value pair
		    else {
			_header.addElement( new KeyValuePair(key, value) );
		    }

		    // Next line
		    line = inputStream.readLine();
		    continue;
		}

		// Field specification
		delimIndex = line.indexOf(":");
		if (delimIndex != -1) {

		    // Get the field identifier
		    String identifier = line.substring(0, delimIndex);
		    identifier = identifier.trim();

		    // Get the field descriptor
		    String descriptor = line.substring(delimIndex+1);
		    descriptor = descriptor.trim();

		    // Add the field
		    _header.addElement( new Field(identifier, descriptor) );

		    // Next line
		    line = inputStream.readLine();
		    continue;
		}

		// Next line
		line = inputStream.readLine();
	    }
	}

	catch (Exception e) {
	    String msg = "Unable to decode NRRD header.";
	    throw new IIOException(msg, e);
	}
    }

    /**
     * Gets the decoded Header.
     *
     * @return Decoded Header.
     */
    public NrrdHeader getHeader()
    {
	return _header;
    }

    /**
     * Determines if the string contains numbers.
     *
     * @param s String.
     *
     * @return True if the string contains numbers; false otherwise.
     */
    private static boolean _isNumber(String s)
    {
	// Check each character
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if ( !Character.isDigit(c) ) { return false; }
	}

	// All numbers
	return true;
    }
}
