/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.blobio.BlobWriter;
import edu.ucla.loni.imageio.DualImageOutputStream;
import edu.ucla.loni.nrrd.NrrdHeader;
import edu.ucla.loni.nrrd.NrrdHeaderEncoder;
import java.awt.image.DataBufferByte;
import java.io.IOException;
import java.util.Vector;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.ImageOutputStream;
import org.w3c.dom.Node;

/**
 * Blob Writer for the NRRD image format.
 *
 * @version 2 November 2010
 */
public class NrrdBlobWriter extends BlobWriter
{
    /** NRRD Header of the blob sequence currently being written. */
    private NrrdHeader _streamHeader;

    /**
     * Constructs an NRRD Blob Writer.
     *
     * @param origProvider Originating provider that instantiated this Object.
     */
    public NrrdBlobWriter(NrrdBlobWriterSpi origProvider)
    {
	super(origProvider);
    }

    /**
     * Sets the output object to the specified Object.  The output destination
     * is assumed to be ready to accept data, and will not be closed at the end
     * of each write. If output is null, any currently set output will be
     * removed.
     *
     * @param output Object to use for future writing.
     *
     * @throws IllegalArgumentException If output is not an instance of one of
     *                                  the classes returned by the originating
     *                                  service provider's getOutputTypes
     *                                  method.
     */
    public void setOutput(Object output)
    {
	// Vector of Image Output Streams
	if (output instanceof Vector) {
	    Vector streams = (Vector)output;

	    // Reset the output if one Image Output Stream
	    if (streams.size() == 1) { output = streams.firstElement(); }

	    // Otherwise convert to a Dual Image Output Stream
	    else { output = new DualImageOutputStream(streams); }
	}

	super.setOutput(output);
    }

    /**
     * Gets an IIOMetadata object containing default values for writing a
     * series of blobs.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultStreamMetadata()
    {
	return new NrrdMetadata(null);
    }

    /**
     * Gets an IIOMetadata object containing default values for writing a
     * blob.
     *
     * @return An IIOMetadata object.
     */
    public IIOMetadata getDefaultBlobMetadata()
    {
	return null;
    }

    /**
     * Prepares an output stream to accept a series of subsequent blob write
     * calls using the provided stream metadata object.
     *
     * @param streamMetadata IIOMetadata object representing stream metadata, or
     *                       null if none is applicable.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set.
     */
    public void prepareWriteBlobSequence(IIOMetadata streamMetadata)
	throws IOException
    {
	// Must have stream metadata
	if (streamMetadata == null) {
	    throw new IOException("No stream metadata.");
	}

	// Construct a stream NRRD Header from the metadata
	try {
	    String rootName = NrrdMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	    Node tree = streamMetadata.getAsTree(rootName);
	    _streamHeader = NrrdMetadataConversions.toNrrdHeader(tree);
	}
      
	catch (Exception e) {
	    String msg="Unable to construct an NRRD Header from the metadata.";
	    throw new IllegalArgumentException(msg);
	}

	// No output has been set
	if (_output == null) {
	    String msg = "No output has been set.";
	    throw new IllegalStateException(msg);
	}

	// Get the image output stream for metadata
	boolean hasTwoStreams = false;
	ImageOutputStream metadataStream;
	if (_output instanceof DualImageOutputStream) {
	    DualImageOutputStream dualStream = (DualImageOutputStream)_output;
	    metadataStream = dualStream.getMetadataOutputStream();
	    hasTwoStreams = true;
	}
	else { metadataStream = (ImageOutputStream)_output; }

	// Write the NRRD header to the metadata stream
	NrrdHeaderEncoder encoder = new NrrdHeaderEncoder();
	encoder.encode(_streamHeader, metadataStream, !hasTwoStreams);
    }

    /**
     * Appends a single blob and possibly associated metadata to the output.
     *
     * @param blobMetadata IIOMetadata object representing blob metadata, or
     *                     null if none is applicable.
     * @param blob DataBuffer containing the binary object.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalArgumentException If the blob is null.
     * @throws IllegalStateException If the output has not been set, or a
     *                               prepareWrite method has not been called.
     */
    public void writeToBlobSequence(IIOMetadata blobMetadata,
				    DataBufferByte blob) throws IOException
    {
	// Check for stream metadata
	if (_streamHeader == null) {
	    String msg = "The blob sequence has not been prepared.";
	    throw new IllegalStateException(msg);
	}

	// Check for a null blob
	if (blob == null) {
	    throw new IllegalStateException("Null blob");
	}

	// Get the image output stream for pixel data
	ImageOutputStream pixelStream;
	if (_output instanceof DualImageOutputStream) {
	    DualImageOutputStream dualStream=(DualImageOutputStream)_output;
	    pixelStream = dualStream.getPixelDataOutputStream();
	}
	else { pixelStream = (ImageOutputStream)_output; }

	// Write the blob data
	pixelStream.write( blob.getData() );
    }

    /**
     * Completes the writing of a series of blobs begun with
     * prepareWriteBlobSequence.  Any blob metadata that should come at the
     * end of the series is written out, and any header information at the
     * beginning of the series is patched up if necessary.
     *
     * @throws IOException If an error occurs during writing.
     * @throws IllegalStateException If the output has not been set, or if
     *                               prepareWriteBlobSequence has not been
     *                               called.
     */
    public void endWriteBlobSequence() throws IOException
    {
	// Reset the metadata
	_streamHeader = null;
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this Blob Writer.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Attempt to close the output stream
	try {

	    // Two Output Streams
	    if (_output instanceof DualImageOutputStream) {
		((DualImageOutputStream)_output).close();
	    }

	    // One Output Stream
	    else if (_output instanceof ImageOutputStream) {
		((ImageOutputStream)_output).close();
	    }
	}
	catch (Exception e) {}
    }
}
