/*
  COPYRIGHT NOTICE
  Copyright (c) 2012  Scott C. Neu and Arthur W. Toga

  This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package edu.ucla.loni.nrrd.plugin;

import edu.ucla.loni.nrrd.Field;
import edu.ucla.loni.nrrd.KeyValuePair;
import edu.ucla.loni.nrrd.KeyValuePairSet;
import edu.ucla.loni.nrrd.NrrdHeader;
import edu.ucla.loni.nrrd.NrrdHeaderElement;
import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadataNode;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Class that provides conversions between an NRRD Header and its DOM Node
 * representation.
 *
 * @version 28 October 2010
 */
public class NrrdMetadataConversions
{
    /** Constructs an NRRD Metadata Conversions. */
    private NrrdMetadataConversions()
    {
    }

    /**
     * Converts the tree to an NRRD Header.
     *
     * @param root DOM Node representing an NRRD Header.
     *
     * @return NRRD Header constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     * @throws IllegalArgumentException If the root Node is null.
     */
    public static NrrdHeader toNrrdHeader(Node root)
	throws IIOInvalidTreeException
    {
	// Check for a null root Node
	if (root == null) {
	    String msg = "A null root Node is not allowed.";
	    throw new IllegalArgumentException(msg);
	}

	// Check that the root Node name matches a required name
	String rootName = NrrdMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	if ( !root.getNodeName().equals(rootName) ) {
	    String msg = "Root node must be named \"" + rootName + "\".";
	    throw new IIOInvalidTreeException(msg, root);
	}

	// Attempt to construct an NRRD Header from the Node
	try {

	    // Get the magic number
	    NamedNodeMap attrs = root.getAttributes();
	    Node magicNode = attrs.getNamedItem(NrrdMetadataFormat.MAGIC);
	    String magicNumber = magicNode.getNodeValue();

	    // Create the header
	    NrrdHeader header = new NrrdHeader(magicNumber);

	    // Convert the child Nodes
	    NodeList childNodes = root.getChildNodes();
	    for (int i = 0; i < childNodes.getLength(); i++) {
		Node childNode = childNodes.item(i);
		String childName = childNode.getNodeName();

		// Convert Fields and add to the header
		if ( NrrdMetadataFormat.FIELDS.equals(childName) ) {
		    Field[] fields = _toFields(childNode);
		    for (int j = 0; j < fields.length; j++) {
			header.addElement(fields[j]);
		    }
		}

		// Convert Key-Value Pairs and add to the header
		else if ( NrrdMetadataFormat.KEYS.equals(childName) ) {
		    KeyValuePair[] pairs = _toPairs(childNode);
		    for (int j = 0; j < pairs.length; j++) {
			header.addElement(pairs[j]);
		    }

		    // Convert the Key-Value Pair Sets
		    NodeList setNodes = childNode.getChildNodes();
		    for (int k = 0; k < setNodes.getLength(); k++) {
			Node setNode = setNodes.item(k);

			// Get the set
			String setName = _toName(setNode.getNodeName());
			KeyValuePairSet pairSet = header.getPairSet(setName);

			// Add the pair
			attrs = setNode.getAttributes();
			Node node = attrs.getNamedItem(NrrdMetadataFormat.KEY);
			String key = node.getNodeValue();
			node = attrs.getNamedItem(NrrdMetadataFormat.VALUE);
			String value = node.getNodeValue();
			pairSet.setPair(key, value);
		    }
		}
	    }

	    // Return the header
	    return header;
	}

	// Unable to construct an NRRD Header
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into an NRRD Header.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the NRRD Header to a tree.
     *
     * @param header NRRD Header to convert.
     *
     * @return DOM Node representing the NRRD Header.
     */
    public static Node toTree(NrrdHeader header)
    {
	// Create the root Node
	String rootName = NrrdMetadataFormat.NATIVE_METADATA_FORMAT_NAME;
	IIOMetadataNode rootNode = new IIOMetadataNode(rootName);
	if (header == null) { return rootNode; }

	// Add the magic number
	rootNode.setAttribute(NrrdMetadataFormat.MAGIC,
			      header.getMagicNumber());

	// Add the fields
	Node fieldsNode = _toFieldTree(header);
	rootNode.appendChild(fieldsNode);

	// Add the keys
	Node keysNode = _toKeyTree(header);
	rootNode.appendChild(keysNode);

	// Add the sets
	NrrdHeaderElement[] elements = header.getElements();
	for (int i = 0; i < elements.length; i++) {
	    if (elements[i] instanceof KeyValuePairSet) {
		KeyValuePairSet set = (KeyValuePairSet)elements[i];
		String name = _toXmlName( set.getName() );

		// Create a Node for each pair
		KeyValuePair[] pairs = set.getPairs();
		for (int j = 0; j < pairs.length; j++) {

		    // Create the Node
		    IIOMetadataNode pairNode = new IIOMetadataNode(name);
		    keysNode.appendChild(pairNode);

		    // Add the pair as attributes
		    pairNode.setAttribute(NrrdMetadataFormat.KEY,
					  pairs[j].getKey());
		    pairNode.setAttribute(NrrdMetadataFormat.VALUE,
					  pairs[j].getValue());
		}
	    }
	}

	// Return the root Node
	return rootNode;
    }

    /**
     * Converts the tree to Fields.
     *
     * @param root DOM Node representing the fields.
     *
     * @return Fields constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     */
    private static Field[] _toFields(Node root) throws IIOInvalidTreeException
    {
	// Attempt to construct Fields from the Node
	try {
	    NamedNodeMap attrs = root.getAttributes();
	    int numOfAttrs = attrs.getLength();

	    // Create a Field for each attribute
	    Field[] fields = new Field[numOfAttrs];
	    for (int i = 0; i < numOfAttrs; i++) {
		Node attr = attrs.item(i);

		// Create the Field
		String identifier = _toName(attr.getNodeName());
		String descriptor = attr.getNodeValue();
		fields[i] = new Field(identifier, descriptor);
	    }

	    // Return the Fields
	    return fields;
	}

	// Unable to construct the fields
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into Fields.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the Fields of the header to a tree.
     *
     * @param header NRRD Header to convert.
     *
     * @return DOM Node representing the fields.
     */
    private static Node _toFieldTree(NrrdHeader header)
    {
	// Create the fields Node
	String name = NrrdMetadataFormat.FIELDS;
	IIOMetadataNode fieldsNode = new IIOMetadataNode(name);

	// Add each field as an attribute
	NrrdHeaderElement[] elements = header.getElements();
	for (int i = 0; i < elements.length; i++) {
	    if (elements[i] instanceof Field) {
		Field field = (Field)elements[i];

		// Encode the attribute name
		name = _toXmlName( field.getIdentifier() );

		// Add the attribute
		fieldsNode.setAttribute(name, field.getDescriptor());
	    }
	}

	// Return the fields Node
	return fieldsNode;
    }

    /**
     * Converts the tree to Key-Value Pairs.
     *
     * @param root DOM Node representing the pairs.
     *
     * @return Key-Value Pairs constructed from the DOM Node.
     *
     * @throws IIOInvalidTreeException If the tree cannot be parsed successfully
     *                                 using the rules of the given format.
     */
    private static KeyValuePair[] _toPairs(Node root)
	throws IIOInvalidTreeException
    {
	// Attempt to construct pairs from the Node
	try {
	    NamedNodeMap attrs = root.getAttributes();
	    int numOfAttrs = attrs.getLength();

	    // Create a pair for each attribute
	    KeyValuePair[] pairs = new KeyValuePair[numOfAttrs];
	    for (int i = 0; i < numOfAttrs; i++) {
		Node attr = attrs.item(i);

		// Create the pair
		String key = _toName(attr.getNodeName());
		String value = attr.getNodeValue();
		pairs[i] = new KeyValuePair(key, value);
	    }

	    // Return the pairs
	    return pairs;
	}

	// Unable to construct the pairs
	catch (Exception e) {
	    String msg = "Cannot convert the DOM Node into Pairs.";
	    throw new IIOInvalidTreeException(msg, e, root);
	}
    }

    /**
     * Converts the Key-Value Pairs of the header to a tree.
     *
     * @param header NRRD Header to convert.
     *
     * @return DOM Node representing the pairs.
     */
    private static Node _toKeyTree(NrrdHeader header)
    {
	// Create the keys Node
	String name = NrrdMetadataFormat.KEYS;
	IIOMetadataNode keysNode = new IIOMetadataNode(name);

	// Add each pair as an attribute
	NrrdHeaderElement[] elements = header.getElements();
	for (int i = 0; i < elements.length; i++) {
	    if (elements[i] instanceof KeyValuePair) {
		KeyValuePair pair = (KeyValuePair)elements[i];

		// Encode the attribute name
		name = _toXmlName( pair.getKey() );

		// Add the attribute
		keysNode.setAttribute(name, pair.getValue());
	    }
	}

	// Return the keys Node
	return keysNode;
    }

    /**
     * Converts the specified legal XML name to its original representation.
     *
     * @param xmlName Legal XML name to convert.
     *
     * @return Original name.
     */
    private static String _toName(String xmlName)
    {
	// -- -> Space
	return xmlName.replaceAll("--", " ");
    }

    /**
     * Converts the specified name into a legal XML name.
     *
     * @param name Name to convert.
     *
     * @return Legal XML name.
     */
    private static String _toXmlName(String name)
    {
	StringBuffer xmlName = new StringBuffer();

	// Encode all special characters
	for (int i = 0; i < name.length(); i++) {
	    char c = name.charAt(i);

	    // Space -> --
	    if (c == ' ') { xmlName.append("--"); }

	    else { xmlName.append(c); }
	}

	return xmlName.toString();
    }
}
